#include "cell.h"
#include <iostream>
using namespace std;

/// Create static variables (set to usable values in main())
double cell::D = 0.0;
double cell::dx = 0.0;
double cell::dy = 0.0;
double cell::k = 0.0;
bool cell::adiabatic = false;
int cell::width = 0;
int cell::height = 0;

cell::cell():
atomic(),
cell_interface(),
du(0.0),
x(0),
y(0)
{
}

void cell::set_pos(int xx, int yy)
{
	x = xx;
	y = yy;
}

void cell::set_init_conditions(double u0[5])
{
	// Initializes the integrator using the supplied data
	// The five points correspond to the four neighboring cells
	// and the value at this cell.
	double delta_u[5];
	du = 0.0;
	q.setParams(D);
	for (int i = 0; i < 5; i++)
	{
		delta_u[i] = u0[i];
	}
	compute_derivative(delta_u);
	q.init(u0[CENTER],du);
}

void cell::init()
{
	hold(q.ta());
}

void cell::delta_int()
{
	// The derivative function computes new values based on changes in
	// the value at this cell and its neighbors.
	double delta_u[5];
	// Since we received no input (this is an internal event), the
	// neighboring values have not changed.
	for (int i = 0; i < 5; i++) delta_u[i] = 0.0;
	// This cell has changed by a single quantum, which we compute by taking
	// the difference of the integrator state (value at last quantum crossing)
	// and integrator output (value at next quantum crossing).
	delta_u[CENTER] = q.out()-q.qstate();
	// Compute the new derivative.
	compute_derivative(delta_u);
	// Perform the integration step.  This sets the value of the last quantum
	// crossing to be equal to the output (i.e., updates the state by a single
	// quantum value).
	q.integ(q.ta(),du);
	// Set the time advance to the next level crossing time.
	hold(q.ta());
}

void cell::delta_ext(stime_t e, const adevs_bag<PortValue>& input)
{
	// Initialize the array of neighboring value changes to zero
	double delta_u[5];
	for (int i = 0; i < 5; i++) delta_u[i] = 0.0;
	// Compute the change at this cell (will be zero if it is really an
	// external event, or equal to a quantum if this is a confluent event).
	delta_u[CENTER] = q.out(e)-q.qstate();
	// Get the value of the changes sent by this cell's neighbors.
	for (int i = 0; i < input.getSize(); i++)
	{
		if (input.get(i).value.x == x-1) delta_u[LEFT] = input.get(i).value.s;
		else if (input.get(i).value.x == x+1) delta_u[RIGHT] = input.get(i).value.s;
		else if (input.get(i).value.y == y-1) delta_u[DOWN] = input.get(i).value.s;
		else if (input.get(i).value.y == y+1) delta_u[UP] = input.get(i).value.s;
	}
	// Compute the new derivatve
	compute_derivative(delta_u);
	// Integrate
	q.integ(e,du);
	// Set the time advance to the time of the next level crossing
	hold(q.ta());
}

void cell::delta_conf(const adevs_bag<PortValue>& input) 
{
	// Activate the external transition function with elapsed time equal to
	// the time advance.
	delta_ext(q.ta(),input);
}

void cell::output_func(adevs_bag<PortValue>& o)
{
	cell_state s;
	// The cell location
	s.x = x;
	s.y = y;
	// Change in the cell state
	s.s = q.out()-q.qstate();
	// Make this value available on the out port of the cell
	output(out,s,o);
}

void cell::gc_output(adevs_bag<PortValue>& g)
{
	// I/O types have automatic scope (see adevs_config.h) so
	// no need to do garbage collection.
}

void cell::report(ofstream& stream) const
{
	stream << q.qstate() << " ";
}

cell::~cell()
{
}

void cell::compute_derivative(double* delta_u)
{
	// Select a value for the edge cells based on boundary conditions
	double edge_cell = 0.0;
	if (adiabatic) edge_cell = delta_u[CENTER];
	// Set boundary conditions at the edge cells
	if (x == 0) delta_u[LEFT] = edge_cell; 
	if (x == width-1) delta_u[RIGHT] = edge_cell; 
	if (y == 0) delta_u[DOWN] = edge_cell; 
	if (y == height-1) delta_u[UP] = edge_cell; 
	// compute new time derivative using center differences in space
	// the new derivative is updated by adding the change in the derivative
	// resulting from the changes in cell states.
	du += (k*(delta_u[LEFT]-2.0*delta_u[CENTER]+delta_u[RIGHT])/(dx*dx) +
				k*(delta_u[UP]-2.0*delta_u[CENTER]+delta_u[DOWN])/(dy*dy));
}
