#ifndef _genr_h_
#define _genr_h_
#include "adevs.h"
#include "job.h"
/*
The genr class produces jobs periodically.  The genr starts
producing jobs when it receives an input on its start port.
It stops producing jobs when it receives an input on its
stop port.  Jobs appear on the out port.
*/
class genr: public atomic 
{
	public:
		/// Constructor.  The generator period is provided here.
		genr(stime_t period):
		atomic(),
		period(period),
		count(0)
		{
		}
		/// State initialization function
		void init()
		{
			// The generator is initially turned on.
			hold(period);
			// Zero jobs produced so far
			count = 0;
		}
		/// Internal transition function
		void delta_int()
		{
			// We just produced a job via the output_func, so increment the
			// job counter
			count = count + 1;
			// Wait until its time to produce the next job
			hold(period);
		}
		/// External transition function
		void delta_ext(stime_t e, const adevs_bag<PortValue>& x)
		{
			// Look for input on the start port.  If input is found,
			// hold until it is time to produce the first output.
			port_iterator start_port(x,start);
			if (!start_port.end_of_input())
			{
				hold(period);
			}
			// Look for input on the stop port.  If input is found,
			// stop the generator by setting our time of next event to infinity.
			port_iterator stop_port(x,stop);
			if (!stop_port.end_of_input())
			{
				passivate();
			}
		}
		/// Confluent transition function
		void delta_conf(const adevs_bag<PortValue>& x)
		{
			// When an internal and external event coincide, compute
			// the internal state transition then process the input.
			delta_int();
			delta_ext(0.0,x);
		}
		/// Output function.  
		void output_func(adevs_bag<PortValue>& y)
		{
			// Place a new job on the output port
			output(out,new job(count),y);
		}
		/// Garbage collection.  This method is called at the
		/// end of every simulation cycle in which the model
		/// produced output.  The bag contains the output objects
		/// produced by the model.
		void gc_output(adevs_bag<PortValue>& g)
		{
			for (int i = 0; i < g.getSize(); i++)
			{
				delete g[i].value;
			}
		}

		/// Model input ports
		static const port_t start;
		static const port_t stop;
		/// Model output port
		static const port_t out;

	private:	

		/// Model state variables
		stime_t period;
		int count;
};

/// Create the static ports and assign them unique 'names' (numbers)
const port_t genr::stop(0);
const port_t genr::start(1);
const port_t genr::out(2);
  
#endif
