#ifndef __proc_h_
#define __proc_h_
#include <stdlib.h>
#include "adevs.h"
#include "job.h"
/*
A processor requires a fixed period of time to service a job.
The processor can serve only one job at a time.  It the processor
is busy, it simply discards incoming jobs.
*/
class proc: public atomic 
{
	public:
		/// Constructor.  The processing time is provided as an argument.
		proc(stime_t proc_time):
		atomic(),
		processing_time(proc_time),
		val(NULL)
		{
		}
		/// State initialization function
		void init()
		{
			// The processor is passive until a job is available to be processed
			passivate();
			if (val != NULL) 
			{
				delete val;
			}
			val = NULL;
		}
		/// Internal transition function
		void delta_int()
		{
			// Done with the job, so set time of next event to infinity
			passivate();
			// Discard the completed job
			if (val != NULL) 
			{
				delete val;
			}
			val = NULL;
		}
		/// External transition function
		void delta_ext(stime_t e, const adevs_bag<PortValue>& x)
		{
			// If we are waiting for a job
			if (timeNext() == ADEVS_INFINITY) 
			{
				// Make a copy of the job (original will be destroyed by the
				// generator at the end of this simulation cycle).
				val = new job(*((job*)x.get(0).value));
				// Wait for the required processing time before outputting the
				// completed job
				hold(processing_time);
			}
			// Otherwise, model just continues with time of next event unchanged
		}
		/// Confluent transition function.
		void delta_conf(const adevs_bag<PortValue>& x)
		{
			// Discard the old job
			delta_int();
			// Process the incoming job
			delta_ext(0.0,x);
		}
		/// Output function.
		void output_func(adevs_bag<PortValue>& y)
		{
			// Produce a copy of the completed job on the out port
			output(out,new job(*val),y);
		}
		/// Garbage collection
		void gc_output(adevs_bag<PortValue>& g)
		{
			for (int i = 0; i < g.getSize(); i++)
			{
				delete g[i].value;
			}
		}
		/// Destructor
		~proc()
		{
			if (val != NULL) 
			{
				delete val;
			}
		}

		/// Model input port
		static const port_t in;
		/// Model output port
		static const port_t out;

	private:	
		/// Model state variables
		stime_t processing_time;
		job* val;
};

/// Create unique 'names' for the model ports.
const port_t proc::in(0);
const port_t proc::out(1);

#endif
