#ifndef __transd_h_
#define __transd_h_
#include <adevs.h>
#include <vector>
#include <iostream>
#include <stdlib.h>
#include "job.h"
/*
The transducer computes various statistics about the
performance of the queuing system.  It receives new jobs
on its ariv port, finished jobs on its solved port,
and generates and output on its out port when the observation
interval has elapsed.
*/
class transd: public atomic 
{
	public:
		/// Constructor
		transd(stime_t observ_time):
		atomic(),
		observation_time(observ_time),
		total_ta(0.0)
		{
		}
		/// State initialization function
		void init()
		{
			total_ta = 0.0;
			jobs_arrived.clear();
			jobs_solved.clear();
			// Collect data until the observation period has ended
			hold(observation_time);
		}
		/// Internal transition function
		void delta_int() 
		{ 
			// Passivate once the data collection period has ended and the
			// results reported.
			passivate(); 
		}
		/// External transition function
		void delta_ext(stime_t e, const adevs_bag<PortValue>& x)
		{
			// Save new jobs in order to compute statistics when they are
			// completed.
			port_iterator ariv_port(x,ariv); 
			for (; !ariv_port.end_of_input(); ariv_port++) 
			{
				job val (*((job*)(*ariv_port)));
				val.t = timeCurrent();
				jobs_arrived.push_back(val);
			}
			// Compute time required to process completed jobs
			port_iterator solved_port(x,solved); 
			for (; !solved_port.end_of_input(); solved_port++) 
			{
				job val(*((job*)(*solved_port)));
				std::vector<job>::iterator i = jobs_arrived.begin();
				for (; i != jobs_arrived.end(); i++)
				{
					if ((*i).value == val.value)
					{
						total_ta += timeCurrent() - (*i).t;
						break;
					}
				}
				val.t = timeCurrent();
				jobs_solved.push_back(val);
			}
		}
		/// Confluent transition function
		void delta_conf(const adevs_bag<PortValue>& x)
		{
			delta_int();
			delta_ext(0.0,x);
		}
		/// Output function
		void output_func(adevs_bag<PortValue>& y)
		{
			/// Model dumps interesting statistics to the console
			/// when the observation period ends.
			stime_t throughput;
			stime_t avg_ta_time;
			if (jobs_solved.size() != 0) 
			{
				avg_ta_time = total_ta / jobs_solved.size();
				if (timeCurrent() > 0)
				{
					throughput = jobs_solved.size() / timeCurrent ();
				}
				else 
				{
					throughput = 0.0;
				}
			}
			else 
			{
				avg_ta_time = 0.0;
				throughput = 0.0;
			}
			std::cout << "End time: " << timeCurrent() << std::endl;
			std::cout << "jobs arrived : " << jobs_arrived.size () << std::endl;
			std::cout << "jobs solved : " << jobs_solved.size() << std::endl;
			std::cout << "AVERAGE TA = " << avg_ta_time << std::endl;
			std::cout << "THROUGHPUT = " << throughput << std::endl;
			/// Generate an output event to stop the generator
			output(out,new object(),y);
		}
		/// Garbage collection
		void gc_output(adevs_bag<PortValue>& g)
		{
			for (int i = 0; i < g.getSize(); i++)
			{
				delete g[i].value;
			}
		}
		~transd()
		{
			jobs_arrived.clear();
			jobs_solved.clear();
		}

		/// Model input port
		static const port_t ariv;
		static const port_t solved;
		/// Model output port
		static const port_t out;

	private:	
		/// Model state variables
		std::vector<job> jobs_arrived;
		std::vector<job> jobs_solved; 
		stime_t observation_time;
		stime_t total_ta;
};

/// Assign unique 'names' to ports
const port_t transd::ariv(0);
const port_t transd::solved(1);
const port_t transd::out(2);

#endif
