#include "phold_network.h"
#include <vector>
#include <iostream>
using namespace std;

phold_network::net_exec::net_exec(int num_nodes,int num_originators,
double net_change_interval,phold_network* parent):
netExec(parent),
r(new crand()),
network_change_interval(net_change_interval)
{
	// Construct the initial set of components
	while (num_nodes > 0)
	{
		// Create a network model 15% of the time
		if (r.uniform(0.0,1.0) < 0.15)
		{
			// Lets not be too picky about accidentally adding an extra model or two
			int net_size = rand()%num_nodes+2;
			int net_orig = 0;
			if (num_originators > 0)
			{
				net_orig = rand()%num_originators;
			}
			// Knock these numbers off of what is left to do
			num_originators -= net_orig;
			num_nodes -= net_size;
			// Add the new network to the component set of this network
			nodes.add(new phold_network(net_size,net_orig,network_change_interval));
		}
		// Otherwise just add a new queue
		else
		{
			num_nodes--;
			nodes.add(new phold_queue(num_originators-- > 0));
		}
	}
	// Connect the network nodes
	build_connections();
}

void phold_network::net_exec::init()
{
	// If the model has a non-infinite structure change interval,
	// then schedule the first structure change.
	if (network_change_interval < ADEVS_INFINITY)
	{
		hold(r.exp(network_change_interval));
	}
	// Otherwise, this will be a static structure model
	else
	{
		passivate();
	}
}

void phold_network::net_exec::delta_int()
{
	// Add and remove some models at random.
	for (int i = 0; i < 2; i++)
	{
		int choice = rand()%4;
		// Remove a node.  This node will be passed to the gc_models()
		// method when it is safe to delete the model.
		if (choice == 0 && nodes.getSize() > 1)
		{
			nodes.remove(0);
		}
		// Add a new network model.  This model will be initialized
		// by the simulation engine.
		else if (choice == 1)
		{
			nodes.add(new phold_network(rand()%5+1,rand()%2));
		}
		// A a new atomic model
		else if (choice == 2)
		{
			nodes.add(new phold_queue(rand()%2 == 0));
		}
	}
	// Rebuild the network topology
	build_connections();
	// Schedule the next structure change
	hold(r.exp(network_change_interval));
}	

void phold_network::net_exec::delta_ext(double e, const adevs_bag<PortValue>& x)
{
	// Our network executive won't respond to inputs
	passivate();
}

void phold_network::net_exec::delta_conf(const adevs_bag<PortValue>& x)
{
	// Our network executive does not respond to inputs
	delta_int();
}

void phold_network::net_exec::output_func(adevs_bag<PortValue>& y)
{
	// Our network executive produces no output
}

void phold_network::net_exec::getComponents(adevs_set<devs*>& c)
{
	// Add all component models (except the network executive and the parent)
	// to the set c.
	c.append(nodes);
}

void phold_network::net_exec::route(const PortValue& pv, devs* model, 
adevs_bag<EventReceiver>& r)
{
	// External inputs go to a randomly selected model
	if (model == getParent())
	{
		EventReceiver rx(nodes.get(rand()%nodes.getSize()),phold_node::in);
		r.add(rx);
	}
	// Otherwise, this is an internal event or external output event
	else
	{
		// Cast the source as a phold_node so that we can access its adj list
		phold_node* src = dynamic_cast<phold_node*>(model);
		// Generate events at each adjacent model
		list<devs*>::iterator i = src->adj.begin();
		for (; i != src->adj.end(); i++)
		{
			// If this is an external output
			if (*i == getParent())
			{
				EventReceiver rx(*i,phold_node::out);
				r.add(rx);
			}
			// Otherwise it is an internal output to input connection
			else
			{
				EventReceiver rx(*i,phold_node::in);
				r.add(rx);
			}
		}
	}
}

void phold_network::net_exec::gc_models(adevs_set<devs*>& removed)
{
	// Delete models that were removed during the last state change
	for (int i = 0; i < removed.getSize(); i++)
	{
		delete removed[i];
	}
}

phold_network::net_exec::~net_exec()
{
	// Delete all of the component models
	for (int i = 0; i < nodes.getSize(); i++)
	{
		delete nodes[i];
	}
}

void phold_network::net_exec::build_connections()
{
	// Build a vector of properly cast nodes
	vector<phold_node*> p_nodes(nodes.getSize());
	// Remove any existing connections
	for (int i = 0; i < nodes.getSize(); i++)
	{
		p_nodes[i] = dynamic_cast<phold_node*>(nodes[i]);
		p_nodes[i]->adj.clear();
	}
	// One model is connected to the external output
	p_nodes[rand()%nodes.getSize()]->adj.push_back(getParent());
	// Setup the random internal network connections.
	for (int i = 0; i < nodes.getSize(); i++)
	{
		int in_degree = rand()%3+1;
		for (int k = 0; k < in_degree; k++)
		{
			int select = rand()%nodes.getSize();
			// Do not connect a model to itself!
			if (p_nodes[select] != p_nodes[i])
			{
				p_nodes[select]->adj.push_back(nodes[i]);
			}
		}
		int out_degree = rand()%3+1;
		for (int k = 0; k < out_degree; k++)
		{
			int select = rand()%nodes.getSize();
			// Do not connect a model to itself!
			if (p_nodes[select] != p_nodes[i])
			{
				p_nodes[i]->adj.push_back(nodes[select]);
			}
		}
	}
	// Done!
}
