#ifndef __phold_network_h_
#define __phold_network_h_
#include "adevs.h"
#include "phold_queue.h"

/**
A phold network consisting of other phold_networks and phold_queues.
*/
class phold_network: public devsn, public phold_node
{
	public:
		/**
		Creates a network with a specified number of atomic nodes, some of which
		can be token originators.  Some of the atomic nodes will be contained
		in a random number of network subcomponents, rather than being contained 
		directly.
		*/
		phold_network(int num_nodes, int num_originators, 
		double net_change_interval = ADEVS_INFINITY):
		devsn(),
		phold_node(),
		nx(num_nodes,num_originators,net_change_interval,this)
		{
		}
		/// Returns the network executive associated with this model
		netExec* getNetExec() { return &nx; }
		/// The destructor destroys all of the component models.
		~phold_network()
		{
		}

	private:
		/**
		The network executive for this network model.  It is defined within
		the context of the network model because it will never exist independently
		of it.
		*/
		class net_exec: public netExec
		{
			public:
				/**
				Create the network using the same arguments as where passed
				to the parent model.  The network executive needs to know
				who its parent is in order to implement external input and output
				connections.
				*/
				net_exec(int num_nodes,int num_originators,
				double network_change_interval, phold_network* parent);
				/// Atomic model methods
				void init();
				void delta_int();
				void delta_ext(double e, const adevs_bag<PortValue>& x);
				void delta_conf(const adevs_bag<PortValue>& x);
				void output_func(adevs_bag<PortValue>& y);
				/// No need for garbage collection because I/O type is primitive
				void gc_output(adevs_bag<PortValue>& g){}
				/**
				These are the network executive specific methods.
				*/
				/**
				Get the set of component models.  This method fills the set c with
				a pointer to every component of the network except the network 
				executive itself and the parent network model.
				*/
				void getComponents(adevs_set<devs*>& c);
				/**
				This is the event routing function.  The method should fill the
				event receiver bag r with EventReceiver objects that describe
				the model and port that needs to receive the pv that was generated
				by the model pointed to by the model argument.
				*/
				void route(const PortValue& pv, devs* model, adevs_bag<EventReceiver>& r);
				/**
				This method is used to garbage collect models which were removed
				from the network during the last simulation cycle.  Changes in the
				model set are computed using the set difference of the model set
				before and after the network executive changes state.  This difference
				is passed the the gc_models() method after those models have been
				removed from the simulation kernel.  At this point, the network
				executive can delete them.
				*/
				void gc_models(adevs_set<devs*>& removed);
				/// Destructor destroys all of the component models
				~net_exec();

			private:
				/// The set of network components
				adevs_set<devs*> nodes;
				/// Random variable for scheduling network restructuring
				rv r;
				double network_change_interval;
				/// Build or rebuild the network
				void build_connections();
		}; // end of the net_exec class

		// Network executive model for the phold_network
		net_exec nx;

};

#endif
