/* godunov.c *************************************************\
*                                                             *
*  Classical first order Godunov scheme for the system        *
*  of conservation laws of 1D gasdynamics,                    *
*  uses exact Newton-type iterative Riemann solver of Godunov *
*                                                             *
*  written by  Andrei A. Chernousov                           *
*  E-mail: <andrei99@iname.com>                               *
*  http://www.geocities.com/andrei_chernousov                 *
*  November 23, 2000                                          * 
**************************************************************/
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <algorithm>
#include "perf_timer.h"

double
   *U1, *U2, *U3, /* conservative variables */
   *F1, *F2, *F3, /* fluxes */

   R, U, P, /* primitive variables */
   R_1, U_1, P_1,
   R_2, U_2, P_2,

   deltaX, deltaT, deltaT_X, Tend;

unsigned LEN, LENN, i, i_, step;
FILE *pF;


/*--- exact Newton-type iterative Riemann solver of Godunov ---*/
#include "solver.c"


/************
*  Array1D  *
************/
double *Array1D( unsigned size )
{
double *x;

   /* C++ */
/* if( ! ( x = new double [size] ) ) {
	  puts( "Cannot allocate memory " );
	  exit( -1 );
   }
   return x;
*/

  /* C */
  if( ( x = (double*)malloc( size * sizeof( double ) ) ) == NULL ) {
	  puts( "Cannot allocate memory " );
	  exit( -1 );
   }
   return x;
   
} /* end Array1D( ) */


/*************
* Initialize *
*************/
void Initialize( void )
{
char str[80];

   /* loading data from file "godunov.ini" */
   if( ( pF = fopen( "godunov.ini", "r" ) ) == NULL ) {
	  printf( "Cannot open file \"godunov.ini\"" );
	  exit( -1 );
   }
   i = 0;
   do{
	  if ( fgets( str, 79, pF ) != NULL ) {
		 switch( i ) {
			case 0:	 LEN    = atoi( str ); break;
			case 1:  deltaX = atof( str ); break;
			case 2:  deltaT = atof( str ); break;
			case 3:  Tend = atof( str ); break;
			default: break;
		 } /* end switch() */
		 i++;
	  } /* end if() */
	  else {
		 puts( "Error while reading file \"godunov.ini\"" );
		 exit( 1 );
	  }
   } while( i < 4 );
   fclose( pF );

   deltaT_X = deltaT / deltaX;
   LENN = LEN+1;

   /* memory allocation */
   U1 = Array1D( LEN+2 ); U2 = Array1D( LEN+2 ); U3 = Array1D( LEN+2 );
   F1 = Array1D( LEN+1 ); F2 = Array1D( LEN+1 ); F3 = Array1D( LEN+1 );

   /* initial conditions */
   for( i = 1, U = 0., R = 2., P = 200000.; i <= LEN/2; i++ ) {
	  U1[i] = R;
	  U2[i] = R * U;
	  U3[i] = P / K_1 + 0.5 * R * U * U;
   } /* end for() */
   for( U = 0., R = 1., P = 100000.; i < LENN; i++ ) {
	  U1[i] = R;
	  U2[i] = R * U;
	  U3[i] = P / K_1 + 0.5 * R * U * U;
   }

} /* end Initialize( ) */


/***********************
* BounCondInGhostCells *
***********************/
void BounCondInGhostCells( void )
{
   /* solid walls */
   U1[0] =   U1[1];
   U2[0] = - U2[1];
   U3[0] =   U3[1];
   U1[LEN+1] =   U1[LEN];
   U2[LEN+1] = - U2[LEN];
   U3[LEN+1] =   U3[LEN];
} /* end BounCondInGhostCells( ) */


/***********
*  Fluxes  *
***********/
void Fluxes( void )
{
   for ( i = 0, i_ = 1; i < LENN; i++, i_++ ) {
	  /* leftmost parameters */
	  R_1 =   U1[i ];
	  U_1 =   U2[i ] / R_1;
	  P_1 = ( U3[i] - 0.5 * R_1 * U_1 * U_1 ) * K_1;
	  /* rightmost parameters */
	  R_2 =   U1[i_];
	  U_2 =   U2[i_] / R_2;
	  P_2 = ( U3[i_] - 0.5 * R_2 * U_2 * U_2 ) * K_1;
	  /* exact Riemann solver of Godunov, see "solver.c" */
	  Godunov( R_1, U_1, P_1,  R_2, U_2, P_2,  &R,  &U,  &P );
	  /* fluxes */
	  F1[i] = R * U;
	  F2[i] = R * U * U + P;
	  F3[i] = U * ( K_K_1 * P + 0.5 * R * U * U );
   }
} /* end Fluxes( ) */


/**************
*  Evolution  *
**************/
void Evolution( void )
{
   for ( i = 1; i < LENN; i++ ) {
	  U1[i] += deltaT_X * ( F1[i-1] - F1[i] );
	  U2[i] += deltaT_X * ( F2[i-1] - F2[i] );
	  U3[i] += deltaT_X * ( F3[i-1] - F3[i] );
   }
} /* end Evolution() */


/***********
*  Output  *
***********/
void Output( void )
{
   /* output to "input.dat" file */
   if ( ( pF = fopen( "input.dat", "w" ) ) == NULL ) {
	  puts( "Cannot open \"input.dat\"" );
	  exit( -1 );
   }
//   fprintf( pF, "0. 90000. 0.9 0. \n" );
   for ( i = 1; i < LENN; i++ ) {
	  R =   U1[i];
	  U =   U2[i] / R;
	  P = ( U3[i] - 0.5 * R * U * U ) * K_1;
	  fprintf( pF,"%3.3f %6.0f %1.4f %3.1f \n", ( i - 0.5 ) * deltaX, P, R, U );
//	  fprintf( pF,"%f %f %f %f \n", ( i - 0.5 ) * deltaX, U1[i], U2[i], U3[i] );
   }
   fclose( pF );
} /* end Output() */


/*********
*  MAIN  *
**********/
int main( void )
{
   /*=== initialization ===*/
   Initialize();
	double t = deltaT; 
	tic();
   /*=== time integration ===*/
   while( t < Tend ) {
	  /*--- advance step number ---*/
	  t += deltaT;
	  /*--- BC in ghost cells ---*/
	  BounCondInGhostCells( );
	  /*--- fluxes at cell interfaces ---*/
	  Fluxes( );
	  /*--- evolution ---*/
	  Evolution( );
   } /* end while */
	printf("%f\n",tocs()); 
   /*--- output of the flowfield ---*/
   Output();
   
   return 0;
   
} /* end main() */

/*--- end godunov.c ---*/
