/* godunov.c *************************************************\
*                                                             *
*  Classical first order Godunov scheme for the system        *
*  of conservation laws of 1D gasdynamics,                    *
*  uses exact Newton-type iterative Riemann solver of Godunov *
*                                                             *
*  written by  Andrei A. Chernousov                           *
*  E-mail: <andrei99@iname.com>                               *
*  http://www.geocities.com/andrei_chernousov                 *
*  November 23, 2000                                          * 
**************************************************************/
#include "params.h"
#include "perf_timer.h"
#include "configuration.h"
#include <fstream>
#include <assert.h>
#include <iostream>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <algorithm>
using namespace std;

static vector<adevs_vec<3> > U; // Conservation variables
static vector<adevs_vec<3> > F; // Fluxes
static double deltaX, deltaT, deltaT_X, Tend; // computation parameters 

/*************
* Initialize *
*************/
static void Initialize(string filename)
{
	configuration config(filename);
	deltaX = config.get_dx();
	deltaT = config.get_dt();
	deltaT_X = deltaT/deltaX;
	Tend = config.get_tend();
	F = U = config.get_data();
} /* end Initialize( ) */


/***********************
* BounCondInGhostCells *
***********************/
static void BounCondInGhostCells()
{
	/* solid walls */
	U[0][0] = U[1][0];
	U[0][1] = -U[1][1];
	U[0][2] = U[1][2];
	U[U.size()-1][0] = U[U.size()-2][0];
	U[U.size()-1][1] = -U[U.size()-2][1];
	U[U.size()-1][2] = U[U.size()-2][2];
} /* end BounCondInGhostCells( ) */


/***********
*  Fluxes  *
***********/
static void Fluxes()
{
	for (unsigned i = 1; i < U.size(); i++)
	{
		F[i-1] = fluxes(U[i-1],U[i]);
	}
} /* end Fluxes( ) */


/**************
*  Evolution  *
**************/
static void Evolution( void )
{
	for (unsigned i = 1; i < U.size()-1; i++ ) 
	{
		U[i] += deltaT_X*(F[i-1]-F[i]);
	}
} /* end Evolution() */


/***********
*  Output  *
***********/
static void Output()
{
	ofstream strm("soln");
	for (unsigned i = 1; i < U.size()-1; i++) 
	{
		double R = U[i][0];
		double UU = U[i][1] / R;
		double P = (U[i][2] - 0.5 * R * UU * UU ) * K_1;
		strm << " " << P << " " << R << " " << UU << endl;
	}
	strm.close();
} /* end Output() */


/*********
*  MAIN  *
**********/
int main(int argc, char** argv)
{
	Initialize(argv[1]);
	double t = 0.0;
	tic();
   /*=== time integration ===*/
   while(t < Tend) {
	  /*--- advance step number ---*/
	  t += deltaT;
	  /*--- BC in ghost cells ---*/
	  BounCondInGhostCells();
	  /*--- fluxes at cell interfaces ---*/
	  Fluxes();
	  /*--- evolution ---*/
	  Evolution();
   } /* end while */
	double e = toc();
	cout << e << " ms" << endl; 
   /*--- output of the flowfield ---*/
   Output();
   return 0;
} /* end main() */
