/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "ball.h"
#include "physics.h"
#include <iostream>

// Ball will only report changes in position in excess of the MOTION_RES value.
const double ball::MOTION_RES = 0.1;

// input and output ports
const port_t ball::collision = 0;
const port_t ball::motion = 1;
const port_t ball::destroyed = 2;
const port_t ball::stop = 3;

ball::ball():atomic(),status(OK){}

ball::ball(const vec& v_init):atomic(),status(OK),v_init(v_init){}

ball::ball(const circle& geom_init):
atomic(),status(OK),geom_init(geom_init),geom(geom_init){}

ball::ball(const circle& geom_init, const vec& v_init):atomic(),
status(OK),v_init(v_init),geom_init(geom_init),geom(geom_init){}

void ball::setWorld(const rectangle& world_size) 
{
	this->world_size = world_size;
}

void ball::setWorld(double x, double y, double width, double height) 
{
	world_size.setPosition(x, y);
	world_size.setSize(width, height);
}

void ball::init() 
{
	status = OK;
	v = v_init;
	geom_op = geom = geom_init;
	dp.x = dp.y = 0;
	hold (0);
}

void ball::delta_int()
{
	// if ball is destroyed, reset and start bouncing again.
	if (status == DESTROYED) 
	{
		geom_op = geom = geom_init;
		v = v_init;
		status = OK;
	}

	// current position gets the last anticipated position
	geom = geom_op;
	// compute the time needed to move a significant distance at
	// the current velocity.
	stime_t dt = timeToDist (MOTION_RES, v);
	// get the position change that will occur in that time
	dp = distInTime (dt, v);
	// make sure we haven't gone off the edge of the world
	checkStatus ();
	// compute the next position
	geom_op.deltaPos (dp);
	// hold for the time needed to move to the next position
	hold (dt);
}

void ball::delta_ext(stime_t e, const adevs_bag<PortValue>& x) 
{
	// update the position of the ball to reflect the elapsed time
	geom.deltaPos (distInTime(e,v));

	// passivate the ball (freeze it) if input is received on the
	// stop port.
	port_iterator stop_port(x,stop);
	if (!stop_port.end_of_input()) 
	{
		passivate();
		return;
	}

	// check for a collision event directed at this ball.
	port_iterator collision_port(x,collision);
	for (; !collision_port.end_of_input(); collision_port++) 
	{
		collisionEvent* cp = (collisionEvent*)(*collision_port);
		// if a collision is detected, compute the balls resulting velocity
		if (cp->getTarget () == this) 
		{
			const shape* s = cp->getGeometry ();
			v = elasticCollision (v, s->surfaceAngle (geom.getPosition ()));
		}
	}

	// compute time to move significant distance at (new?) velocity
	stime_t dt = timeToDist (MOTION_RES, v);
	// compute change in position to occur in that time
	dp = distInTime (dt, v);
	// compute the next position
	geom_op = geom;
	geom_op.deltaPos (dp);
	// hold until position is reached 
	hold (dt);
}

void ball::delta_conf(const adevs_bag<PortValue>& x) 
{ 
	delta_int();
	delta_ext(ta(),x); 
}

void ball::output_func(adevs_bag<PortValue>& y) 
{
	// if destroyed, output the balls new status
	if (status != OK)
		output(destroyed,new statusEvent(this, status, &geom),y);
	// output the balls updated position
	output(motion, new motionEvent(this, v, &geom_op),y);
}

// destroy the ball if it goes outside of the bounding box
void ball::checkStatus() 
{
	if (!world_size.overlaps (geom)) status = DESTROYED;
}

void ball::gc_output(adevs_bag<PortValue>& g)
{
	for (int i = 0; i < g.getSize(); i++)
		delete g[i].value;
}
