/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "block.h"
#include <iostream>
#include <stdlib.h>
using namespace std;

// input and output ports
const port_t block::status = 1;
const port_t block::ball_collision = 2;
const port_t block::paddle_collision = 3;
const port_t block::paddle_motion = 4;
const port_t block::ball_motion = 5;

block::block():atomic(),fixed(false){}

block::block(const rectangle& geom_init):
atomic(),fixed(false),geom_init(geom_init),geom(geom_init){}

block::block(const rectangle& geom_init, eStatus stat):
atomic(),fixed(true),geom_init(geom_init),geom(geom_init),
stat_init(stat),stat(stat){}

void block::init()
{
	// need to report our status at the beginning.  Afterwards,
	// only report if block is destroyed.
	report_status = true;
	// clean out the collision lists
	ball_collisions.clear();
	paddle_collisions.clear();

	geom = geom_init;
	if (fixed) stat = stat_init;

	// if status is not fixed, select OK or INVULNERABLE at random
	else if (random() < (RAND_MAX / 4)) stat = INVULNERABLE;
	else stat = OK;
	hold (0);
}

void block::delta_int() 
{
	// clear the report_status falg
	report_status = false;
	// clear out the collision lists
	ball_collisions.clear();
	paddle_collisions.clear();
	passivate();
}

void block::delta_ext(stime_t e, const adevs_bag<PortValue>& x) 
{
	// if block is destroyed, passivate and don't respond to any more
	// events.
	if (stat == DESTROYED) 
	{
		passivate();
		return;
	}

	// check for a collision with a ball 
	port_iterator ball_motion_port(x,ball_motion);
	for (; !ball_motion_port.end_of_input(); ball_motion_port++) 
	{
		motionEvent* e = (motionEvent*)(*ball_motion_port);
		circle* ball_geom = (circle*)e->getPosition ();
		// if a collision is detected, add a collision event to the 
		// ball collision list and (if appropriate) set the status
		// of the block to DESTROYED.
		if (geom.overlaps (*ball_geom)) 
		{
			ball_collisions.push_back(new collisionEvent(e->getSource(),&geom)); 
			if (stat != INVULNERABLE) 
			{
				report_status = true;
				stat = DESTROYED;
			}
		}
	}

	// check for a collision with a paddle
	port_iterator paddle_motion_port(x,paddle_motion);
	for (; !paddle_motion_port.end_of_input(); paddle_motion_port++) 
	{
		motionEvent* e = (motionEvent*)(*paddle_motion_port);
		// if a collision is detected, add a collision event
		// to the paddle collision list
		if (geom.overlaps (*(rectangle*)e->getPosition ())) 
		{
			paddle_collisions.push_back(new collisionEvent (e->getSource(), &geom)); 
		}
	}

	// generate output immediately
	hold (0);
}

void block::output_func(adevs_bag<PortValue>& y) 
{
	// report the status if it has changed or on a reset
	if (report_status)
		output(status, new statusEvent(this, stat, &geom), y);

	list<object*>::iterator i;
	// output all of the detected collisions
	for (i = ball_collisions.begin(); i != ball_collisions.end(); i++) 
		output(ball_collision, *i, y); 
	for (i = paddle_collisions.begin(); i != paddle_collisions.end(); i++) 
		output(paddle_collision, *i, y); 
}

void block::delta_conf(const adevs_bag<PortValue>& x) 
{
	delta_int();
	delta_ext(0,x);
}

void block::gc_output(adevs_bag<PortValue>& g)
{
	for (int i = 0; i < g.getSize(); i++)
		delete g[i].value;
}
