/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "paddle.h"
#include "physics.h"
#include "events.h"
#include <iostream>

// input and output ports
const port_t paddle::collision = 0;
const port_t paddle::vcontrol = 1;
const port_t paddle::ball_motion = 2;
const port_t paddle::motion = 3;
const port_t paddle::ball_collision = 4;
const port_t paddle::stop = 5;

// motion resolution (distance considered significant)
const double paddle::MOTION_RES = 0.5;

// zero velocity vec
const vec paddle::V_ZERO (0, 0);

paddle::paddle():
atomic(),v_paddle(1.5,0){}

paddle::paddle(const rectangle& geom_init):
atomic(),v_paddle(1.5,0),geom_init(geom_init),geom(geom_init),geom_op(geom_init){}

paddle::paddle(const vec& v_paddle):
atomic(),v_paddle(v_paddle){}

paddle::paddle (const rectangle& geom_init, const vec& v_paddle):atomic(),
v_paddle(v_paddle),geom_init(geom_init),geom(geom_init),geom_op(geom_init){}

void paddle::init()
{
	collisions.clear();
	geom = geom_op = geom_init;
	v = V_ZERO;
	vec dp = V_ZERO;
	hold (0);
}

void paddle::delta_int() 
{
	// clear the collision list
	collisions.clear();
	// update the current position of the paddle
	geom = geom_op;
	// compute the next position of the paddle
	vec dp = distInTime (timeToDist (MOTION_RES, v), v);
	geom_op.deltaPos (dp);
	// hold until the next position is reached
	hold (timeToDist (MOTION_RES, v));
}

void paddle::delta_ext(stime_t e, const adevs_bag<PortValue>& x) 
{
	// update the position of the paddle to reflect the elapsed time
	geom.deltaPos (distInTime (elapsed (), v));
	geom_op = geom;

	// passivate if a stop message is received. 
	port_iterator stop_port(x,stop);
	if (!stop_port.end_of_input()) 
	{
		passivate ();
		return;
	}

	bool collision_occuring = false;

	// check for collisions with a wall
	port_iterator collision_port(x,collision);
	if (!collision_port.end_of_input())
	{ 
		collision_occuring = true;
		collisionEvent* e = (collisionEvent*)(*collision_port);
		const rectangle* r = (rectangle*)e->getGeometry ();
		// back out of the collision area
		vec vback;
		if (r->getPosition().x <= geom.getPosition().x) vback = v_paddle;
		else vback = -1 * v_paddle;
		while (geom.overlaps (*r)) 
			geom.deltaPos (vback);
		geom_op = geom;
		// stop the paddle
		v = V_ZERO;
		hold (0);
	}

	// check for control signals (skip this if stopping because of a collision)
	port_iterator vcontrol_port(x,vcontrol);
	if (!vcontrol_port.end_of_input() && !collision_occuring) 
	{
		// update the paddle velocity 
		paddleControlEvent* e = (paddleControlEvent*)(*vcontrol_port);
		if (e->getControl () == LEFT) v = v_paddle * -1;
		else if (e->getControl () == RIGHT) v = v_paddle;
		else v = V_ZERO;
		hold (0);
	}

	// check for ball collision
	port_iterator ball_motion_port(x,ball_motion);
	for (; !ball_motion_port.end_of_input(); ball_motion_port++) 
	{
		motionEvent* e = (motionEvent*)(*ball_motion_port);
		// if a collision occured, add the collision event to the collision list
		if (geom.overlaps (*(circle*)e->getPosition ())) 
		{
			hold (0);
			collisions.push_back(new collisionEvent (e->getSource(), &geom)); 
		}
	}

	// unless holding for an immediate output, compute next position
	// and hold for the time required to reach it.
	if (ta () > 0) 
	{
		dp = distInTime (sigma(), v);
		geom_op.deltaPos (dp);
		hold (sigma ());
	}
}

void paddle::delta_conf(const adevs_bag<PortValue>& x) 
{
	delta_int();
	delta_ext(ta(),x);
}

void paddle::output_func(adevs_bag<PortValue>& y) 
{
	// output new position
	output (motion, new motionEvent(this, &geom), y);
	// output all collision events
	std::list<object*>::iterator i = collisions.begin();
	for (; i != collisions.end(); i++) 
		output(ball_collision, *i, y); 
}

void paddle::gc_output(adevs_bag<PortValue>& g) 
{
	for (int i = 0; i < g.getSize(); i++)
		delete g[i].value;
}
