/**********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "viewer.h"
#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <stdio.h>
using namespace std;

// class for holding old positions
class __pos
{
	public:
		bool first;
		int x, y, width, height;
		__pos():first(true),x(0),y(0),width(0),height(0){}
		__pos(const __pos& src):first(src.first),x(src.x),y(src.y),
		width(src.width),height(src.height){}
		const __pos& operator=(const __pos& src)
		{
			x = src.x; y = src.y; width = src.width; height = src.height;
			first = src.first;
			return *this;
		}
		~__pos(){}

};

// input and output ports
const port_t viewer::ball_motion = 0;
const port_t viewer::block_position = 1;
const port_t viewer::paddle_motion = 2;
const port_t viewer::paddle_control = 3;
const port_t viewer::ball_status = 4;
const port_t viewer::stop = 5;

// period with which the input queue is checked
const stime_t viewer::PERIOD = 1.0;

/**
The code here is standard for X windows applications.
Does some parsing of the command line looking for display
options.
*/
viewer::viewer (char** argv, int argc, int width, int height):
atomic() 
{
	// open the display
	display = XOpenDisplay (NULL);

	if (display == NULL) 
	{
		cout << "xblock: Could not connect to display." << endl;
		exit (-1);
	}
	screen = DefaultScreen (display);
	// get screen dimensions
	// create the window at (0,0)
	win = XCreateSimpleWindow (display, XRootWindow (display, screen),
			0, 0, width, height, 0, 
			XBlackPixel (display, screen),
			XBlackPixel (display, screen));
	// set the size hints.  Defaults for all other properties.
	size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
	size_hints.x = size_hints.y = 0;
	size_hints.width = size_hints.min_width = size_hints.max_width = width;
	size_hints.height = size_hints.min_height = size_hints.max_height = height;

	XSetWMProperties (display, win, NULL, NULL, argv, argc,
			&size_hints, NULL, NULL);

	memset (&event_mask, 0xff, sizeof (long));
	XSelectInput (display, win, KeyPressMask | StructureNotifyMask | KeyReleaseMask);
	// xor graphics contexts.
	grphc_cntxt = XCreateGC (display, XRootWindow (display, screen), 0, NULL);
	XSetFunction (display, grphc_cntxt, GXxor);
	XSetForeground (display, grphc_cntxt, XWhitePixel (display, screen));
	XSetBackground (display, grphc_cntxt, XBlackPixel (display, screen));
	// make the window visible
	XMapWindow (display, win);
	XClearWindow (display, win);
	XAutoRepeatOff (display);
	// initial number of balls and the score
	balls = 2;
	score = 0;
}

void viewer::init() 
{
	quit = false;
	block_count = 0;
	XClearWindow (display, win);
	checkEvents ();
	XSync (display, False);
	if (control_event != NULL) delete control_event;
	control_event = NULL;
	pos_history.clear ();
	drawScore(score, true);
	drawBallCount(balls, true);
	hold (PERIOD);
}

void viewer::delta_int() 
{
	// clear the last control output
	control_event = NULL;
	// if game is over, passivate
	if (quit || (block_count == 0 && timeCurrent () > PERIOD))
		passivate ();
	// otherwise hold until the next user input period
	else 
		hold (PERIOD);
	// check the event queue
	checkEvents ();
}

void viewer::delta_ext(stime_t e, const adevs_bag<PortValue>& x) 
{
	// clear and then draw the ball position
	port_iterator ball_motion_port(x,ball_motion);
	for (; !ball_motion_port.end_of_input(); ball_motion_port++) 
	{
		motionEvent* e = (motionEvent*)(*ball_motion_port);;
		const circle* c = (circle*)e->getPosition();
		drawBall(*c,pos_history[e->getSource()]);
	}

	// clear and then draw the block position
	port_iterator block_position_port(x,block_position);
	for (; !block_position_port.end_of_input(); block_position_port++) 
	{
		statusEvent* e = (statusEvent*)(*block_position_port);
		const rectangle* r = (rectangle*)e->getGeometry();
		// update the score and remaining block count if the block was destroyed
		if (e->getStatus() == DESTROYED) 
		{
			score++;
			block_count--;
			drawScore(score);
		}
		// otherwise, if block is destroyable increment the block count
		else if (e->getStatus() == OK) 
			block_count++;
		drawBlock(*r,e->getStatus());
	}

	// clear and the draw the paddle
	port_iterator paddle_motion_port(x,paddle_motion);
	for (; !paddle_motion_port.end_of_input(); paddle_motion_port++) 
	{
		motionEvent* e = (motionEvent*)(*paddle_motion_port);
		const rectangle* r = (rectangle*)e->getPosition ();
		drawPaddle(*r,pos_history[e->getSource()]); 
	}

	// check the ball status
	port_iterator ball_status_port(x,ball_status);
	for (; !ball_status_port.end_of_input(); ball_status_port++) 
	{
		statusEvent* e = (statusEvent*)(*ball_status_port);
		// if the ball is destroyed, decrement the ball count.
		// if it is zero, end the game
		if (e->getStatus () == DESTROYED) 
		{
			balls--;
			drawBallCount (balls);
			if (balls < 0) 
			{
				quit = true;
				hold (0);
			}
		}
	}
}

void viewer::delta_conf(const adevs_bag<PortValue>& x) 
{ 
	delta_int();
	delta_ext(0.0,x); 
}

void viewer::output_func(adevs_bag<PortValue>& y) 
{ 
	// output the paddle control event
	if (control_event != NULL) 
		output(paddle_control, control_event, y);
	// output a stop event if the game is over
	if (quit || (block_count == 0 && timeCurrent () > PERIOD))
		output(stop, new intg(), y);
}

void viewer::gc_output(adevs_bag<PortValue>& g)
{
	for (int i = 0; i < g.getSize(); i++)
		delete g[i].value;
}
 
viewer::~viewer()
{
	XFreeGC (display, grphc_cntxt);
	XAutoRepeatOn (display);
	XCloseDisplay (display);
	pos_history.clear ();
}

void viewer::drawScore(int score, bool first) 
{
	if (!first) 
		XDrawString (display, win, grphc_cntxt, 10, 10, last_score, strlen (last_score));
	sprintf (last_score, "Score: %d", score);
	XDrawString (display, win, grphc_cntxt, 10, 10, last_score, strlen (last_score));
}
 
void viewer::drawBallCount(int balls, bool first) 
{
	if (!first) 
		XDrawString(display,win,grphc_cntxt,10,25,last_ball_count,strlen(last_ball_count));
	sprintf (last_ball_count, "Balls: %d", balls);
	XDrawString(display,win,grphc_cntxt,10,25,last_ball_count,strlen(last_ball_count));
}

void viewer::drawPaddle(const rectangle& r, __pos& old)
{
	int _x = (int)rint(r.getPosition().x);
	int _y = (int)rint(r.getPosition().y);
	int _w = (int)rint(r.getWidth());
	int _h = (int)rint(r.getHeight());
	if (!old.first) 
		XDrawRectangle (display, win, grphc_cntxt, old.x, old.y, old.width, old.height);
	else old.first = false;
	old.x = _x;
	old.y = _y;
	old.width = _w;
	old.height = _h;
	XDrawRectangle (display, win, grphc_cntxt, _x, _y, _w, _h);
}

void viewer::drawBall(const circle& c, __pos& old)
{
	int _x = (int)rint(c.getPosition().x - c.getRadius() / 2.0);
	int _y = (int)rint(c.getPosition().y - c.getRadius() / 2.0); 
	int _radius = (int)rint(c.getRadius() * 2.0); 
	if (!old.first) 
		XDrawArc(display, win, grphc_cntxt, old.x, old.y, old.width, old.height, 
		0, 360*64);
	else old.first = false;
	old.x = _x;
	old.y = _y;
	old.width = old.height = _radius;
	XDrawArc (display, win, grphc_cntxt, _x, _y, _radius, _radius, 0, 360*64);
}

void viewer::drawBlock(const rectangle& r, eStatus status)
{
	int _x = (int)rint(r.getPosition().x);
	int _y = (int)rint(r.getPosition().y);
	int _w = (int)rint(r.getWidth());
	int _h = (int)rint(r.getHeight());
	XDrawRectangle (display, win, grphc_cntxt, _x, _y, _w, _h);
	if (status != INVULNERABLE)
		XFillRectangle (display, win, grphc_cntxt, _x, _y, _w, _h);
}

void viewer::checkEvents() 
{ 
	int vk;
	while (XCheckMaskEvent (display, event_mask, &event) == True) 
	{
		if (event.type == KeyPress) 
		{
			vk = XKeycodeToKeysym (display, event.xkey.keycode, 0);
			if (control_event != NULL) delete control_event;
			if (vk == XK_Right) control_event = new paddleControlEvent (RIGHT);
			else if (vk == XK_Left) control_event = new paddleControlEvent (LEFT);
			else if (vk == XK_Q || vk == XK_q) quit = true;
		}
		else if (event.type == KeyRelease) 
		{
			vk = XKeycodeToKeysym (display, event.xkey.keycode, 0);
			if (control_event != NULL) delete control_event;
			if (vk == XK_Right || vk == XK_Left) 
				control_event = new paddleControlEvent (STOP);
		} 
	}
}
