/***********
Copyright (C) 1999, 2000, 2002  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "cell.h"
#include <assert.h>
using namespace std;

// Define the cell's ports.
const port_t cell::u(0);
const port_t cell::n(1);
const port_t cell::q(2);
const port_t cell::b(3);

// All cells share a common display object.  This is set externally.
viewer* cell::view(NULL);

cell::cell (const point& location, bool alive)
:atomic (),loc(location),alive(alive),nalive(0){}

void cell::init()
{
	// If we are a new cell with > 0 neighbors, decide if we are alive
	if (nalive > 0) alive = (nalive == 3);
	// if we are alive, then should output our state at the next timestep
	if (alive) hold(1);
	// otherwise, we're dead so just sit there.
	else passivate();
}

/**
Internal transition function.
*/
void cell::delta_int() 
{
	// assume we are going to passivate.  Can reschedule later if necessary.
	passivate();
	// if we are alive but doomed due to the number of living neighbors
	// then change phase to dead and hold (hold time in this case was chosen
	// arbitrarily)
	if (alive && !canLive()) 
	{
		alive = false; 
		hold(1);
	}
}

/**
External transition function.
*/
void cell::delta_ext(stime_t e, const adevs_bag<PortValue>& x) 
{
	// get the bag of messages on the input port.
	port_iterator nport(x,n);
	// for each message in the bag, increment our neighbor
	// count by the value of the message (+1 if a neighbor
	// was born, -1 if just deceased).
	for (; !nport.end_of_input(); nport++) 
	{
		nalive += VALUE((*nport));
	}
	// if we have exactly three neighbors and are currently dead, then we are born.
	// Hold so that we can inform our neighbors of our new state.
	if (nalive == 3 && !alive) 
	{
		alive = true;
		hold(1);
	}

	// otherwise, if we are doomed due to our neighbor count,
	// set our phase to dead and hold.
	else if(!canLive() && alive) 
	{
		alive = false;
		hold(1);
	}

	// otherwise, just sit there.
	else 
		passivate();

}

/**
Confluent transition function.
*/
void cell::delta_conf(const adevs_bag<PortValue>& x) 
{
	delta_ext(0,x);
}

/**
Output function.
*/
void cell::output_func(adevs_bag<PortValue>& y) 
{
	// tell the viewer our current state.
	view->setState(loc, alive);
	// if we are alive, output a 1 and a birth message
	if (alive) 
	{
		assert (nalive == 3 || nalive == 0);
		output(u,MAKE_VALUE(1),y);
		output(b,loc,y); 
	}
	// else output a -1 and our position
	else output(u,MAKE_VALUE(-1),y);
	// if we have zero neighbors and are not alive, output our position
	if (!alive && nalive == 0) output(q,loc,y);
}

cell::~cell(){}
