/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <adevs.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include "point.h"
#include "viewer.h"
#include "cell.h"
#include "cellspace.h"
#include <assert.h>
#include <iostream>
#include <vector>
using namespace std;

#define NSIZE 8

const port_t cellspace::birthed(1);
const port_t cellspace::isolated(0);

cellspace::cellspace(int width, int height):
netExec(NULL),width(width),height(height){}

/**
Reset/initialization function.
*/
void cellspace::init()
{
	// Remove old cells
	map<point,cell*>::iterator i = cells.begin();
	for (; i != cells.end(); i++)
	{
		delete (*i).second;
	}
	cells.clear();
	srand(clock());
	for (int x = 0; x < width; x++)
	{
		for (int y = 0; y < width; y++)
		{
			point p(map_point(x,y));
			if (cells.count(p) == 0 && rand() % 4 == 0)
				cells[p] = new cell(p,true);
		}
	}
	passivate();
}

void cellspace::delta_int() { passivate(); }
void cellspace::delta_conf(const adevs_bag<PortValue>& x) { delta_ext(0,x); }
void cellspace::delta_ext(stime_t e, const adevs_bag<PortValue>& x) 
{
	// remove the isolated cells
	port_iterator isolated_port(x,isolated);
	for (; !isolated_port.end_of_input(); isolated_port++) 
	{
		cells.erase(*isolated_port);
	}

	// add the cells who have living neighbors
	map<point,cell*> new_cells;
	point l[NSIZE];
	// for every cell that was birthed
	port_iterator birthed_port(x,birthed);
	for (; !birthed_port.end_of_input(); birthed_port++) 
	{
		// get its location
		point p(*birthed_port);
		// compute its neighbors
		getAdj(p,l);
		// for each neighbor
		for (int j = 0; j < NSIZE; j++) 
		{
			// if this neighbor is not already in the cellspace  
			if (cells.count(l[j]) == 0) 
			{
				// if this neighbor is not already in the list of new cells
				if (new_cells.count(l[j]) == 0) 
				{
					// add the cell to the new cell list 
					new_cells[l[j]] = new cell(l[j],false);
				}
				// increment the living neighbor count for the cell 
				new_cells[l[j]]->incrAlive();
			}
		}
	}
	// add the new cells
	for (map<point,cell*>::iterator i = new_cells.begin(); i != new_cells.end(); i++)
		cells[(*i).first] = (*i).second;
	passivate();
}

void cellspace::output_func(adevs_bag<PortValue>& y){}

void cellspace::getComponents(adevs_set<devs*>& c)
{
	// Remove everything from the cellspace
	map<point,cell*>::iterator i = cells.begin();
	for (; i != cells.end(); i++)
		c.add((*i).second);
}

/**
routing function
*/
void cellspace::route(const PortValue& pv, devs* src, adevs_bag<EventReceiver>& r) 
{
	// If a cell was isolated, send that output to this model
	if (pv.port == cell::q)
		r.add(EventReceiver(this,isolated));
	// If a cell was birthed, send that output to this model
	else if (pv.port == cell::b)
		r.add(EventReceiver(this,birthed));
	// otherwise, report the +/-1 neighbor to all of the src's neighbors
	else 
	{
		// cast the source of the event to a cell and get its location.
		cell* c = (cell*)src;
		// compute the Moore neightborhood, adding an EventReceiver (model, port)
		// pair to the neightbor list for each adjacent cell.  If a non-existing cell is
		// to going to be the recipient of an event, that cell will be created
		// and added to the simulator with the appropriate neighbor count during
		// the external transition.
		point l[NSIZE];
		getAdj(c->getLocation(), l);
		for (int i = 0; i < NSIZE; i++) 
		{
			if (cells.count(l[i]) == 1)
			{
				r.add(EventReceiver(cells[l[i]], cell::n));
			}
		}
	}
}

/**
Fills the array of NSIZE points with the neighbors of p
*/
void cellspace::getAdj(const point& p, point* l)
{
	point n;
	int offset = 0;
	for (int i = -1; i < 2; i++) 
	{
		X(n) = X(p) + i; Y(n) = Y(p) - 1;
		l[offset] = map_point(X(n),Y(n));
		offset++;
	}
	for (int i = -1; i < 2; i++) 
	{
		X(n) = X(p) + i; Y(n) = Y(p) + 1;
		l[offset] = map_point(X(n),Y(n));
		offset++;
	}
	X(n) = X(p) + 1; Y(n) = Y(p);
	l[offset] = map_point(X(n),Y(n));
	offset++;
	X(n) = X(p) - 1; Y(n) = Y(p);
	l[offset] = map_point(X(n),Y(n));
}

/**
Maps the point (x,y) into the cell's torrodial space.
*/
point cellspace::map_point(int x,int y) const 
{
	int nx = x;
	int ny = y;
	if (x < 0) nx = width - 1;
	else if (x >= width) nx = 0;
	if (y < 0) ny = height - 1;
	else if (y >= height) ny = 0;
	return MAKE_POINT(nx,ny);
}

cellspace::~cellspace() 
{
	// Remove everything from the cellspace
	map<point,cell*>::iterator i = cells.begin();
	for (; i != cells.end(); i++)
		delete (*i).second;
}
