/***********
Copyright (C) 1999, 2000, 2002  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "viewer.h"
#include <string>
#include <iostream>
#include <math.h>
#include <stdlib.h>
using namespace std;

/**
The code here is standard for X windows applications.
Does some parsing of the command line looking for display
options.
*/

const int viewer::MINDIM = 20;
const int viewer::DEF_DIM = 25;

void viewer::parseCommand (const char* cmd) 
{
	unsigned i;
	string arg = cmd;
	char value [100];

	size_t loc = arg.find ("-s", 0, strlen ("-s"));
	if (loc != string::npos) screensaver = true;

	loc = arg.find ("-root", 0, strlen ("-root"));
	if (loc != string::npos) 
	{
		root = true;
		screensaver = false;
	}

	loc = arg.find ("-G", 0, strlen ("-G"));
	if (loc == string::npos) return;

	int j = 0;
	for (i = strlen ("-G"); i < arg.length () && isdigit (cmd[i]); i++) 
		value[j++] = cmd[i];   
	value[j] = (char)NULL;
	if (strlen(value) > 0)
		gheight = atoi (value);

	j = 0;
	i++;
	for (;i < arg.length () && isdigit (cmd[i]); i++) 
		value[j++] = cmd[i];   
	value[j] = (char)NULL;
	if (strlen(value) > 0)
		gwidth = atoi (value);
}

viewer::viewer(char** argv, int argc) 
{
	// open the display
	display = XOpenDisplay (NULL);
	if (display == NULL) 
	{
		adevs_exception err(string("xlife: Could not connect to ") + XDisplayName(NULL));
		throw err;
	}

	gheight = gwidth = DEF_DIM;
	screensaver = false;

	for (int i = 1; i < argc; i++)
		parseCommand (argv[i]);

	screen = DefaultScreen (display);

	// get screen dimensions
	if (!screensaver && !root) 
	{
		dwidth = DisplayWidth (display, screen) / 2;
		dheight = DisplayHeight (display, screen) / 2;
		if (gheight < MINDIM || gheight > dheight) gheight = dheight;
		if (gwidth < MINDIM || gwidth > dwidth) gwidth = dwidth;
	}

	else 
	{
		dwidth = DisplayWidth (display, screen);
		dheight = DisplayHeight (display, screen);
		if (gheight < MINDIM || gheight > dheight) gheight = dheight;
		if (gwidth < MINDIM || gwidth > dwidth) gwidth = dwidth;
	}

	xscale = (double)dwidth / (double)gwidth;
	yscale = (double)dheight / (double)gheight;
	cellw = (int)ceil(xscale);
	cellh = (int)ceil(yscale);

	// create the window at (0,0).  Window will fill the entire
	// display.
	if (screensaver) 
	{
		XSetWindowAttributes attr;
		attr.override_redirect = True;
		win = XCreateWindow (display, XRootWindow (display, screen),
			0, 0, dwidth, dheight, 0, CopyFromParent,
			InputOutput, CopyFromParent, CWOverrideRedirect, &attr);
	}
	else if (root) 
		win = XRootWindow (display, screen);
	else 
	{
		win = XCreateSimpleWindow (display, XRootWindow (display, screen),
			0, 0, dwidth, dheight, 0, 
			XBlackPixel (display, screen),
			XBlackPixel (display, screen));
		// set the size hints.  Defaults for all other properties.
		size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
		size_hints.x = size_hints.y = 0;
		size_hints.width = size_hints.min_width = size_hints.max_width = dwidth;
		size_hints.height = size_hints.min_height = size_hints.max_height = dheight;
		XSetWMProperties (display, win, NULL, NULL, argv, argc,
			&size_hints, NULL, NULL);
	}

	// respond to key presses to quit.  
	if (!root) 
	{
		memset (&event_mask, 0xff, sizeof (long));
		XSelectInput (display, win, KeyPressMask);
	}

	// black and white graphics contexts.
	deadgc = XCreateGC (display, XRootWindow (display, screen), 0, NULL);
	livegc = XCreateGC (display, XRootWindow (display, screen), 0, NULL);
	XSetForeground (display, deadgc, XBlackPixel (display, screen));
	XSetForeground (display, livegc, XWhitePixel (display, screen));
	XSetBackground (display, deadgc, XBlackPixel (display, screen));
	XSetBackground (display, livegc, XBlackPixel (display, screen));

	// make the window visible
	XMapWindow (display, win);
	XClearWindow (display, win);

	if (screensaver) 
	{
		XGrabServer (display);
		XGrabPointer (display, win, False, PointerMotionMask | ButtonPressMask |
			ButtonReleaseMask, GrabModeAsync,
			GrabModeAsync, None, None, CurrentTime);
		XGrabKeyboard (display, win, False, GrabModeAsync,
			GrabModeAsync, CurrentTime);
	}
}

void viewer::setState(const point& p, bool alive) 
{
	int x = (int)rint((double)X(p) * xscale);
	int y = (int)rint((double)Y(p) * yscale); 
	if (alive) XFillRectangle(display, win, livegc, x, y, cellw, cellh); 
	else XFillRectangle(display, win, deadgc, x, y, cellw, cellh); 
}

int viewer::getWidth() const { return gwidth; }
int viewer::getHeight() const { return gheight; }

void viewer::blank() 
{
	XClearWindow (display, win);
}

bool viewer::checkEvents()
{ 
	bool results = false;
	while (XCheckMaskEvent (display, event_mask, &event) == True) 
	{
		if (!screensaver)
			results = event.type == KeyPress;
		else
			results = true;
	}
	return results;
}

viewer::~viewer()
{
	if (screensaver) 
	{
		XUngrabServer (display);
		XUngrabKeyboard (display, CurrentTime);
		XUngrabPointer (display, CurrentTime);
	}
	XFreeGC (display, deadgc);
	XFreeGC (display, livegc);
	XCloseDisplay (display);
}
