/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001, 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __adevs_hashtable_h_
#define __adevs_hashtable_h_
#include "adevs_config.h"
#include "adevs_bag.h"
#include "adevs_func.h"
#include <stdlib.h>
#include <vector>
#include <iostream>
#define _COMPUTE_ADEVS_HASH_FUNC_(key,modulus) ((unsigned)hash_function(key))%modulus

ADEVS_NAMESPACE_BEGIN

/**
Function for finding prime table sizes.
Uses a lookup table to find precomupted prime numbers.
k is the index into the table.
*/
int get_prime(int k);

/**
A hashtable.  Key equivalence is defined by the == operator.
Keys and values must support the equality and assignment operators.
The hash_func must be a class that supports the function operator()
where the function operator accepts a const reference to a key and returns an
integer hashcode.
*/
template <class key_t,class value_t,class hash_func=id<key_t> > class adevs_hashtable
{
	public:
		/**
		Default constructor.  Creates an empty hashtable.  If the min_cap
		value is supplied, the hashtable is guaranteed not to need a rehash
		until at least that many elements have been inserted.
		*/
		adevs_hashtable(int min_cap = 0):
		keys(0),
		prime_table_index(0)
		{
			while (get_prime(prime_table_index)*2 < min_cap)
			{
				prime_table_index++;
			}
			buckets = get_prime(prime_table_index);
			avail = buckets;
			table = new bucket<key_t,value_t>[buckets];
		}
		/**
		Copy constructor.  Keys and values are copied using the assignment
		operator.
		*/
		adevs_hashtable(const adevs_hashtable<key_t,value_t,hash_func>& src):
		buckets(src.buckets),
		keys(src.keys),
		prime_table_index(src.prime_table_index),
		avail(src.avail)
		{
			table = new bucket<key_t,value_t>[buckets];
			for (int i = 0; i < buckets; i++)
			{
				table[i] = src.table[i];
			}
		}
		/**
		Assignment operator.  End result is the same as would have been achieved
		using the copy constructor.
		*/
		const adevs_hashtable<key_t,value_t,hash_func>& 
		operator=(const adevs_hashtable<key_t,value_t,hash_func>& src)
		{
			if (src.buckets != buckets)
			{
				delete [] table;
				table = new bucket<key_t,value_t>[src.buckets];
			}
			prime_table_index = src.prime_table_index;
			keys = src.keys;
			avail = src.avail;
			buckets = src.buckets;
			for (int i = 0; i < buckets; i++)
			{
				table[i] = src.table[i];
			}
			return *this;
		}
		/**
		Add an item.  This replaces the existing item with that key.
		*/
		void insert(const key_t& key, const value_t& value)
		{
			int pos = find_key(key);
			if (table[pos].empty) 
			{
				avail--;
				keys++;
			}
			else if (table[pos].deleted)
			{
				keys++;
			}
			table[pos].deleted = false;
			table[pos].empty = false;
			table[pos].key = key;
			table[pos].value = value;
			if (avail <= buckets / 2) 
			{
				rehash();
			}
		} 
		/**
		Remove the item associated with the supplied key.
		*/
		void remove(const key_t& key)
		{
			int pos = find_key(key);
			if (!table[pos].empty) 
			{
				keys--;
				table[pos].deleted = true;
			}
		}
		/**
		Returns the value associated with key or inserts a new value if key not found
		The new value is constructed using the default (no argument) constructor.
		*/
		value_t& operator[](const key_t& key)
		{
			int pos = find_key(key);
			if (table[pos].empty || table[pos].deleted) 
			{
				value_t default_value;
				insert(key,default_value);
				pos = find_key(key);
			}
			return table[pos].value;
		}
		/**
		Returns a constant reference to the value associated with the supplied
		key.
		*/
		const value_t& find(const key_t &key)
		{
			return (*this)[key];
		}
		/**
		Looks for key and assigns its associated value to the value argument if it is
		found.  The method returns true if the key was found and false otherwise.
		*/
		bool find(const key_t& key, value_t& value)
		{
			int pos = find_key(key);
			bool found = (!table[pos].empty && !table[pos].deleted);
			if (found)
			{
				value = table[pos].value;
			}
			return found;
		}
		/// Returns true if the key exists in the hashtable, false otherwise.
		bool contains(const key_t& key) const
		{
			int pos = find_key(key);
			return (!table[pos].empty && !table[pos].deleted);
		}
		/// Gives the number of keys in the hashtable.
		int getSize() const { return keys; }
		/**
		Removes all keys and values.
		*/
		void clear()
		{
			for (int i = 0; i < buckets; i++)
			{
				table[i].empty = true;
				table[i].deleted = false;
			}
			avail = buckets;
			keys = 0;
		}
		/// Get the key,value pairs in this hashtable.
		std::vector<std::pair<key_t,value_t> > getElements() const
		{
			std::vector<std::pair<key_t,value_t> > els;
			int key_count = 0;
			for (int i = 0; key_count < keys; i++)
			{
				if (!table[i].empty && !table[i].deleted)
				{
					els.push_back(std::pair<key_t,value_t>(table[i].key,table[i].value));
					key_count++;
				}
			}
			return els;
		}
		/// Returns the number of hashing buckets that are available in the table.
		int getBucketCount() const { return buckets; }
		/// Destructor.
		~adevs_hashtable()
		{
			delete [] table;
		}

	private:	
		template <class _key_t, class _value_t> class bucket
		{
			public:
				_key_t key;
				_value_t value;
				bool empty;
				bool deleted;
				bucket():empty(true),deleted(false){}
				bucket(const _key_t& key, const _value_t& value):
				key(key),value(value),empty(false),deleted(false){}
				bucket(const bucket<_key_t,_value_t>& src):
				key(src.key),value(src.value),empty(src.empty),deleted(src.deleted){}
				const bucket<_key_t,_value_t>& 
				operator=(const bucket<_key_t,_value_t>& src)
				{
					key = src.key;
					value = src.value;
					empty = src.empty;
					deleted = src.deleted;
					return *this;
				}
				~bucket(){}
		};
		int buckets;
		bucket<key_t,value_t>* table;
		int keys;
		int prime_table_index;
		int avail;
		hash_func hash_function;

		int find_key(const key_t& key) const
		{
			int index = _COMPUTE_ADEVS_HASH_FUNC_(key,buckets);
			int offset = 0;
			while (!table[index].empty && (table[index].key != key))
			{
				index += 2*(++offset)-1;
				if (index >= buckets) index -= buckets;
			}
			return index;
		}

		void rehash()
		{
			int key_count = keys;
			bucket<key_t,value_t>* key_value_pairs = table;
			keys = 0;
			if (key_count >= buckets/2)
			{
				prime_table_index++;
			}
			buckets = get_prime(prime_table_index);
			avail = buckets;
			table = new bucket<key_t,value_t>[buckets];
			for (int i = 0; keys < key_count; i++)
			{
				if (!key_value_pairs[i].empty && !key_value_pairs[i].deleted)
				{
					insert(key_value_pairs[i].key,key_value_pairs[i].value);
				}
			}
			delete [] key_value_pairs;
		}

};

ADEVS_NAMESPACE_END

#endif
