/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2003-2005 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __adevs_models_h_
#define __adevs_models_h_
#include "adevs_config.h"
#include "adevs_bag.h"
#include "adevs_set.h"
#include "adevs_exceptions.h"
#include "adevs_types.h"
#include <algorithm>
#include <cmath>

ADEVS_NAMESPACE_BEGIN

/**
Model I/O is represented by PortValue pairs.
*/
class PortValue
{
	public:
		/// The value of this PortValue pair.
		ADEVS_IO_TYPE value;
		/// The port on which the value appears.
		ADEVS_PORT_TYPE port;
		/// Default constructor.
		PortValue()
		{
		}
		/// Creates an object with the indicated value on the indicated port.
		PortValue(ADEVS_PORT_TYPE port, ADEVS_IO_TYPE value):
		value(value),
		port(port)
		{
		}
		/// Copy constructor.
		PortValue(const PortValue& src):
		value(src.value),
		port(src.port)
		{
		}
		/// Assignment operator.
		const PortValue& operator=(const PortValue& src)
		{
			value = src.value;
			port = src.port;
			return *this;
		}
		/// Destructor.
		~PortValue()
		{
		}
};

/**
Predeclare network and atomic model so types can be used as the type of
parent in the basic devs model and for type ID functions.  
*/
class devsn;
class atomic;
class netExec;

/**
This structure provides basic simulator support that is needed
for all types (network and atomic) of models.
*/
struct basic_sim_support_t
{
	// Preferred thread assignment for NUMA computers.  A -1 indicated no preference.
	static int preferred_thread;
	/// Pointer the network model that has this model as a component
	devsn* parent;
	/// An integer key used to keep track of the model location in a cellpsace
	unsigned long key;
};
	
/**
The devs class provides basic operations for all devs models.
Note that models are derived from the object class.  Therefore
(if using the default configuration) models can, themselves,
be passed as I/O between models (this can be useful when building
some types of dynamic structure models).
*/
class devs: public object
{
	public:
		/// Default constructor.
		devs():
		object()
		{
			bsim_support.parent = NULL;
			bsim_support.key = 0;
		}
		/// Copy constructor.
		devs(const devs& src):
		object(src)
		{
			bsim_support.parent = NULL;
			bsim_support.key = src.bsim_support.key;
		}
		/// Assignment operator.
		void operator=(const devs& src)
		{
			object::operator=(src);
			bsim_support.parent = NULL;
			bsim_support.key = src.bsim_support.key;
		}
		/**
		The init function should place the model into its initial state.
		The time t0 is the initial simulation time.
		*/
		virtual void initialize(ADEVS_TIME_TYPE t0) = 0;
		/// Destructor.
		virtual ~devs()
		{
		}
		/**
		Returns NULL if this is not a network model and returns a pointer to
		itself otherwise.  This method is used to avoid a relatively expensive
		dynamic cast.
		*/
		virtual devsn* typeIsNetwork() { return NULL; }
		/// Returns NULL if this is not a atomic model; returns itself otherwise.
		virtual atomic* typeIsAtomic() { return NULL; }
		/// Returns NULL if this is not a network executive; returns itself otherwise.
		virtual netExec* typeIsNetExec() { return NULL; }
		/**
		Get the model that contains this model as a component.  Returns
		NULL if this model is at the top of the hierarchy.
		*/
		const devsn* getParent() const { return bsim_support.parent; }
		devsn* getParent() { return bsim_support.parent; }
		/**
		Assign a new parent to this model.  This method is intended to
		support copy constructors and assignment operators for end user
		network models.
		*/
		void setParent(devsn* parent) { bsim_support.parent = parent; }
		/**
		Set the preferred thread assignment for model constructed subsequent to this call.  
		Calling this method forces subsequent memory allocations to occur in the
		memory block associated with the specified thread.
		Atomic models created after this call will be assigned
		to that same thread.  This method assumes that all thread creation is done serially!
		The method is not thread safe and should only be used when model creation is done
		serially, but simulator execution will be done in parallel.  Moreover, the method
		only has an effect on NUMA machines and only if ADEVS_USE_LINUX_NUMA is defined
		in the adevs_config file.
		*/
		static void prefer_thread(int thread_id);
		/**
		Simulator support data.  This should not be used in any way for
		system modeling.
		*/
		basic_sim_support_t bsim_support;
};

/**
EventReceiver objects are used to route I/O within a coupled model
(see the netExec class).
*/
class EventReceiver
{
	public:
		/// The destination model.
		devs* model;
		/// The destination port.
		ADEVS_PORT_TYPE port;
		/// Default constructor.  Model pointer is set to NULL.
		EventReceiver():
		model(NULL)
		{
		}
		/// Create an event receiver that points to the provided model and port.
		EventReceiver(devs* model, ADEVS_PORT_TYPE port):
		model(model),
		port(port)
		{
		}
		/// Copy constructor.
		EventReceiver(const EventReceiver& src):
		model(src.model),
		port(src.port)
		{
		}
		/// Assignment operator.
		const EventReceiver& operator=(const EventReceiver& src)
		{
			model = src.model;
			port = src.port;
			return *this;
		}
		/// Two EventReceivers are equal if they point to the same model and port.
		bool operator==(const EventReceiver& other) const
		{
			return (model == other.model && port == other.port);
		}
		/// Inequality operator.
		bool operator!=(const EventReceiver& other) const
		{
			return !((*this) == other);
		}
		/// Destructor.
		~EventReceiver()
		{
		}
};

/**
Hash function for event receiver objecsts.
*/
struct EventReceiverHashFunc
{
	unsigned long operator()(const EventReceiver& rx) const
	{
		return (unsigned long)(rx.model);
	}
};
		
/**
The port_iterator can be used to iterator over the values on a specific port
in an input bag (i.e., adevs_bag<PortValue> types).
*/
class port_iterator
{
	public:
		/// Copy constructor.
		port_iterator(const port_iterator& src):
		x(src.x),
		pos(src.pos)
		{
		}
		/**
		Create an iterator for bag x that will iterate over values on the 
		indicated port.
		*/
		port_iterator(const adevs_bag<PortValue>& x, ADEVS_PORT_TYPE port):
		x(&x),
		pos(0),
		port(port)
		{
			while (pos < x.getSize())
			{
				if (x.get(pos).port == port) return;
				else pos++;
			}
		}
		/// Advance the iterator to the next value on the port.
		void operator++() 
		{ 
			if (pos < 0) pos = 0;
			else pos++;
			while (pos < x->getSize())
			{
				if (x->get(pos).port == port) return;
				else pos++;
			}
		}
		/// Advance the iterator to the next value on the port.
		void operator++(int i)
		{
			operator++();
		}
		/// Return the iterator to the previous value on the port.
		void operator--()
		{
			if (pos < 0) pos = 0;
			else pos--;
			while (pos >= 0)
			{
				if (x->get(pos).port == port) return;
				else pos--;
			}
		}
		/// Return the iterator to the previous value on the port.
		void operator--(int i)
		{
			operator--();
		}
		/// Get the current value of the iterator (that is, the value on the port).
		ADEVS_IO_TYPE operator*() const { return (x->get(pos)).value; }
		/// Returns true if at the end of the input on the port.
		bool end_of_input() const { return pos >= x->getSize(); }
		/// Returns true if at the beginning of the input on the port.
		bool start_of_input() const { return pos < 0; }
		/// Destructor.
		~port_iterator()
		{
		}
	private:
		const adevs_bag<PortValue>* x;
		int pos;
		ADEVS_PORT_TYPE port;
		void operator=(const port_iterator& src){}
		port_iterator(){}
};

/**
Simulator support structure for atomic models.  The most generic simulation
support is included in the devs class, which atomic is derived from.
*/
struct atomic_sim_support_t
{
	// Pointer to the sim_thread object that is responsible for this model
	void* owning_thr;
	// Preferred thread assignment for this model
	int preferred_thread;
	// Current simulation time
	ADEVS_TIME_TYPE t;
	// Time of last event
	ADEVS_TIME_TYPE tL;
	// Time of next event
	ADEVS_TIME_TYPE tN;
	// Schedule position index (see the adevs_sched class)
	int q_index;
	// Flag used to mark model as active in a simulation cycle
	bool active;
	// Bag of input values for this simulation cycle
	adevs_bag<PortValue>* x;
	// Bag of output values for this simulation cycle
	adevs_bag<PortValue>* y;
};

/**
Base type for all atomic DEVS models.
*/
class atomic: public devs
{
	public:
		/// Constructor.
		atomic():
		devs()
		{
			// Preferred thread assignment set to match last user specified preference (see devs class)
			asim_support.preferred_thread = bsim_support.preferred_thread;
			// Rest of values get defaults
			asim_support.owning_thr = NULL;
			asim_support.tN = hold_value = ADEVS_INFINITY;
			asim_support.tL = ADEVS_ZERO;
			asim_support.t = ADEVS_ZERO;
			asim_support.q_index = -1;
			asim_support.active = false;
			asim_support.x = NULL;
			asim_support.y = NULL;
		}
		/// Copy constructor.
		atomic(const atomic& src):
		devs(src)
		{
			asim_support.preferred_thread = src.asim_support.preferred_thread;
			asim_support.owning_thr = src.asim_support.owning_thr;
			hold_value = src.hold_value;
			asim_support.tN = src.asim_support.tN;
			asim_support.tL = src.asim_support.tL;
			asim_support.t = src.asim_support.t;
			asim_support.q_index = -1;
			asim_support.active = false;
			asim_support.x = NULL;
			asim_support.y = NULL;
		}
		/// Assignment operator.
		const atomic& operator=(const atomic& src)
		{
			devs::operator=(src);
			asim_support.preferred_thread = src.asim_support.preferred_thread;
			asim_support.owning_thr = src.asim_support.owning_thr;
			hold_value = src.hold_value;
			asim_support.tN = src.asim_support.tN;
			asim_support.tL = src.asim_support.tL;
			asim_support.t = src.asim_support.t;
			asim_support.q_index = -1;
			asim_support.active = false;
			asim_support.x = asim_support.y = NULL;
			return *this;
		}
		/// Calls the init() method with timeCurrent() equal to t0.
		void initialize(ADEVS_TIME_TYPE t0)
		{
			init();
		}
		/// State initialization function.
		virtual void init() = 0;
		/// Internal transition function.
		virtual void delta_int() = 0;
		/// External transition function.
		virtual void delta_ext(ADEVS_TIME_TYPE e, const adevs_bag<PortValue>& x) = 0;
		/// Confluent transition function.
		virtual void delta_conf(const adevs_bag<PortValue>& x) = 0;
		/// Output function.  Output values should be added to the bag y.
		virtual void output_func(adevs_bag<PortValue>& y) = 0;
		/**
		Garbage collection function.  The PortValue objects in g are
		no longer in use by the simulation engine and should be disposed of. 
`		Note that the elements in g are only those PortValue objects produced as
		output by this model.
		*/
		virtual void gc_output(adevs_bag<PortValue>& g) = 0;
		/// Destructor.
		virtual ~atomic()
		{
		}
		/// Adds the PortValue pair (port,value) to the bag y.
		static void output(ADEVS_PORT_TYPE port, ADEVS_IO_TYPE value, 
		adevs_bag<PortValue>& y)
		{
			y.add(PortValue(port,value));
		}
		/**
		Get the time of the next internal event. If the ta() method is
		overriden, then this method will only return the time of next
		event that was valid when the model was activated (e.g., during
		an internal event it will always return a value equal to timeCurrent()).
		*/
		ADEVS_TIME_TYPE timeNext() const { return asim_support.tN; }
		/// Get the time of the last state transition.
		ADEVS_TIME_TYPE timeLast() const { return asim_support.tL; }
		/// Get the current simulation time.
		ADEVS_TIME_TYPE timeCurrent() const { return asim_support.t; }
		/**
		Get the time until the next event.  This returns timeNext()-timeCurrent(),
		and so the same restrictions apply here as with timeNext() if the
		ta() method has been overridden.
		*/
		ADEVS_TIME_TYPE sigma() const { return asim_support.tN-asim_support.t; }
		/// Get the time since the last event
		ADEVS_TIME_TYPE elapsed() const { return asim_support.t-asim_support.tL; }
		/**
		Get the value of the time advance function.  The default implementation
		returns the last value passed to the hold() method.  If this method is
		overriden, then the hold() and passivate() methods can not be used, and the
		timeNext() and sigma() values will not immediately reflect the time advance.
		*/
		virtual ADEVS_TIME_TYPE ta() const { return hold_value; }
		/**
		Sets the time of next event (timeNext()) to timeCurrent()+dt.
		Throws an adevs_exception if dt < ADEVS_ZERO.  This is the preferred method
		for scheduling an event.  It can not be used in conjunction with an overriden
		ta() method.
		*/
		void hold(ADEVS_TIME_TYPE dt) 
		{ 
			if (dt >= ADEVS_ZERO) 
			{ 
				hold_value = dt;
				if (dt < ADEVS_INFINITY) asim_support.tN = asim_support.t + dt;
				else asim_support.tN = ADEVS_INFINITY;	
			}
			else THROW_BASE_EXCEPTION("time advance < ADEVS_ZERO")
		}
		/**
		Set the time of next event to ADEVS_INFINITY.  This can not be used in conjunction with
		an overriden ta() method.
		*/
		void passivate() { hold_value = ADEVS_INFINITY; }
		/// Returns a pointer to this model.
		atomic* typeIsAtomic() { return this; }
		/**
		This structure is used to support the simulation engine.  It should not
		be used in any way by a system modeler.
		*/
		atomic_sim_support_t asim_support;

	private:
		/**
		Hold value of the default time advance function.
		*/
		ADEVS_TIME_TYPE hold_value;
};

/**
Simulation support structure that is specific to network executives.
*/
struct net_exec_sim_support_t
{
	// Garbage collection set for models removed during a structure change
	adevs_set<devs*> garbage;
	// Set of network executives that have changed state and are safe to delete
	adevs_set<netExec*>* safe_set;
};

/**
Base type for all network executive models.
*/
class netExec: public atomic
{
	public:
		/**
		Constructor. The parent should be the devsn model to which this network
		executive belongs.
		*/
		netExec(devsn* parent = NULL):
		atomic()
		{
			nsim_support.safe_set = NULL;
			bsim_support.parent = parent;
		}
		/**
		Copy constructor.  The parent should be the parent of the copy to be made,
		which made be different from the parent of the original.
		*/
		netExec(const netExec& src, devsn* parent):
		atomic(src)
		{
			nsim_support.safe_set = NULL;
			bsim_support.parent = parent;
		}
		/**
		Implementations of this method should fill the
		set c with all components models, excluding the parent model
		and the network executive itself.
		*/
		virtual void getComponents(adevs_set<devs*>& c) = 0;
		/**
		An implementation should fill the EventReceiver bag r
		with all destinations to receive pv.value.
		*/
		virtual void route(const PortValue& pv, devs* model, 
		adevs_bag<EventReceiver>& r) = 0;
		/**
		The models in the set removed are no longer members of this
		network executive's component set (as per getComponents())
		and are no longer in use by the simulation engine.  They should
		be properly disposed of by the network executive.
		*/
		virtual void gc_models(adevs_set<devs*>& removed) = 0;
		/**
		This method returns a hint to the simulation engine. If the
		return value is true, the simulation engine can skip the check
		for new or removed components.  Otherwise, the simulator will
		perform the new/removed component check.  By default, this method
		will always return true.
		*/
		virtual bool componentSetChanged() const
		{
			return true;
		}
		/// Destructor.
		virtual ~netExec()
		{
			// If we have access to a safe to delete set, make sure
			// we're removed from it
			if (nsim_support.safe_set != NULL)
			{
				nsim_support.safe_set->erase(this);
			}
		}
		/**
		Returns a pointer to this model.
		*/
		netExec* typeIsNetExec() { return this; }
		/**
		This structure is used to support the simulation engine.  It should not
		be used in any way by a system modeler.
		*/
		net_exec_sim_support_t nsim_support;

	protected:
		/// Copy constructor.
		netExec(const netExec& src):
		atomic(src)
		{
			nsim_support.safe_set = NULL;
		}
		/// Assignment operator.
		void operator=(const netExec& src)
		{
			atomic::operator=(src);
			nsim_support.safe_set = NULL;
		}
};

/**
Base class for DEVS network models.
*/
class devsn: public devs
{
	public:
		/// Constructor.
		devsn():
		devs()
		{
		}
		/// Copy constructor.
		devsn(const devsn& src):
		devs(src)
		{
		}
		/// Assignment operator.
		const devsn& operator=(const devsn& src)
		{
			devs::operator=(src);
			return *this;
		}
		/**
		Any class derived from the devsn class should call
		the parent init() method when the derived init() method
		is called.  The init() method initializes all
		component models and sets their parent to point at this model.
		The init() method assumes that the network executive
		has already been created (that is, getNetExec() != NULL).
		*/
		void initialize(ADEVS_TIME_TYPE t0)
		{
			// Initialize the network executive
			getNetExec()->bsim_support.parent = this;
			getNetExec()->asim_support.t = t0;
			getNetExec()->asim_support.tL = t0;
			getNetExec()->initialize(t0);
			// Initialize the component models
			adevs_set<devs*> models;
			getNetExec()->getComponents(models);
			for (int i = 0; i < models.getSize(); i++)
			{
				models[i]->bsim_support.parent = this;
				atomic* amodel = models[i]->typeIsAtomic();
				if (amodel != NULL)
				{
					amodel->asim_support.t = t0;
					amodel->asim_support.tL = t0;
				}
				models[i]->initialize(t0);
			}
		}
		/// Get the network executive associated with this network model.
		virtual netExec* getNetExec() = 0;
		/**
		Destructor.  This destructor does not delete any component models.
		Any cleanup should be done by the derived class.
		*/
		virtual ~devsn()
		{
		}
		/// Returns a pointer to this model.
		devsn* typeIsNetwork() { return this; }
};

ADEVS_NAMESPACE_END

#endif
