/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2005 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __adevs_pqueue_h_
#define __adevs_pqueue_h_
#include "adevs_config.h"
#include "adevs_models.h"
#define SCHED_DEFAULT_CAP 20

ADEVS_NAMESPACE_BEGIN

/**
A scheduler for atomic models.  The scheduler is implemented as a binary
heap.  The scheduler requires that the time types be able to support all
of the comparison operations, assignment operators, and copy constructors.
*/
class adevs_sched
{
	public:
		/**
		Creates a scheduler with the default or specified initial capacity.
		*/
		adevs_sched(int capacity = SCHED_DEFAULT_CAP);
		/**
		Add a model to the schedule or change to position of a model already in
		the schedule. If the tN is ADEVS_INFINITY, the the item or removed
		(if it is already in the queue) or not inserted.  Because this method
		uses the model's q_index field, it is imperative that a model be 
		entered into one, and only one, schedule at any time.
		*/
		void schedule(atomic* model, ADEVS_TIME_TYPE tN);
		/**
		Get the model at the front of the queue.  
		Returns NULL if the queue is empty.
		*/
		atomic* getMinimum() const { return heap[1].item; }
		/**
		Remove the model at the front of the queue.
		*/
		void removeMinimum();
		/**
		Returns the smallest time of next event, or ADEVS_INFINITY if
		the queue is empty.
		*/
		ADEVS_TIME_TYPE minPriority() const { return heap[1].priority; }
		/// Returns the number of items in the queue.
		int getSize() const { return size; }
		/// Returns true if the queue is empty, and false otherwise.
		bool isEmpty() const { return size == 0; }
		/// Removes all items from the queue.
		void clear();
		/// Destructor.
		~adevs_sched();

	private:
		// Definition of an element in the heap.
		class heap_element 
		{
			public:
				atomic* item;
				ADEVS_TIME_TYPE priority;
				heap_element():
				item(NULL),
				priority(ADEVS_INFINITY)
				{
				}
				heap_element(atomic* item,ADEVS_TIME_TYPE priority):
				item(item),
				priority(priority)
				{
				}
				heap_element(const heap_element& src):
				item(src.item),
				priority(src.priority)
				{
				}
				const heap_element& operator=(const heap_element& src)
				{
					item = src.item;
					priority = src.priority;
					return *this;
				}
				~heap_element()
				{
				}
		};
		heap_element* heap;
		int capacity;
		int size;

		/// No copy constructor
		adevs_sched(const adevs_sched& src){}
		/// No assignment operator
		void operator=(const adevs_sched& src){}

		/// Double the schedule capacity
		void enlarge();
		/**
		Heap percolation functions.  Returns the empty slot that emerges for
		a new item with the provided priority.
		*/
		int percolate_down(int index, ADEVS_TIME_TYPE priority);
		int percolate_up(int index, ADEVS_TIME_TYPE priority);
};

ADEVS_NAMESPACE_END

#undef SCHED_DEFAULT_CAP 
#endif
