/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999-2005 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __adevs_set_h_
#define __adevs_set_h_
#include "adevs_config.h"
#include "adevs_bag.h"
#include "adevs_hashtable.h"

ADEVS_NAMESPACE_BEGIN

/**
A set can contain a single instance of an element.
A hashing function must be provided that operates on the
element type (default is to use the id() function).
*/
template <class element_t,class hash_func=id<element_t> > class adevs_set: 
public adevs_bag<element_t> 
{
	public:
		/// Default constructor.  Creates an empty set.
		adevs_set():
		adevs_bag<element_t>(),
		el_table()
		{
		}
		/// Creates a set with a specified initial capacity and capacity increment value.
		adevs_set(int capacity):
		adevs_bag<element_t>(capacity),
		el_table(capacity)
		{
		}
		/**
		Copy constructor.  Elements are copied using the assignment operator and
		copy constructor.
		*/
		adevs_set(const adevs_set<element_t,hash_func>& src):
		adevs_bag<element_t>(src),
		el_table(src.el_table)
		{
		}
		/**
		Assignment operator.  Elements are copied using the assignment operator and
		copy constructor.
		*/
		const adevs_set<element_t,hash_func>& 
		operator=(const adevs_set<element_t,hash_func>& src)
		{
			adevs_bag<element_t>::operator=(src);
			el_table = src.el_table;
			return *this;
		}
		/// Adds the element to the set if it is not already a member of the set.
		void add(const element_t& e)
		{
			if (!el_table.contains(e))
			{
				adevs_bag<element_t>::add(e);
				el_table.insert(e,e);
			}
		}
		/// Removes the element at the specified index.
		void remove(int index)
		{
			el_table.remove((*this)[index]);
			adevs_bag<element_t>::remove(index);
		}
		/// Removes all of the elements from the set.
		void removeAll()
		{
			while (this->getSize() > 0)
				remove(0);
		}
		/**
		Returns true if the element is in the set, and false otherwise.  Equality
		is determined using the == operator.
		*/
		bool contains(const element_t& e) const 
		{
			return el_table.contains(e);
		}
		/// Set difference operator.
		adevs_set<element_t,hash_func> 
		operator-(const adevs_set<element_t,hash_func>& x) const 
		{
			adevs_set<element_t,hash_func> result(*this);
			int size = result.getSize();
			for (int i = size-1; i >= 0; i--)
			{
				if (x.contains(result[i]))
					result.remove(i);
			}
			return result;
		}
		/// Compute the intersection of this and another set.
		adevs_set<element_t,hash_func>
		intersection(const adevs_set<element_t,hash_func>& x) const
		{
			adevs_set<element_t,hash_func> result;
			for (int i = 0; i < this->getSize(); i++)
			{
				if (x.contains(this->get(i)))
				{
					result.add(this->get(i));
				}
			}
			return result;
		}
		/// Destructor.
		~adevs_set(){}
	private:
		adevs_hashtable<element_t,element_t,hash_func> el_table;
};

ADEVS_NAMESPACE_END

#endif
