/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __adevs_types_h_
#define __adevs_types_h_
#include "adevs_config.h"
#include "adevs_exceptions.h"
#include <string>

ADEVS_NAMESPACE_BEGIN

#ifndef NO_PORT_T_TYPE
typedef ADEVS_PORT_TYPE port_t;
#endif
typedef ADEVS_TIME_TYPE stime_t;

ADEVS_NAMESPACE_END

/**
The object class is used as the basic I/O type in the default configuration.
*/
class object
{
	public:
		/// Constructor.
		object(){}
		/// Copy constructor.
		object(const object&){}
		/// Assignment operator.
		const object& operator=(const object&)
		{
			return *this;
		}
		/**
		The clone() method should return a copy of this object,
		or throw an adevs_exception() if the clone() method is
		not supported (this is the default behavior).
		*/
		virtual object* clone() const
		{
			THROW_BASE_EXCEPTION("This object does not support clone()")
		}
		/// Destructor.
		virtual ~object(){}
};

/**
An integer type wrapped in an object.
*/
class intg: public object 
{
	public:	
		int value;
		/// Creates an intg with initial value of 0.
		intg():object(),value(0){}
		/// Copy constructor. 
		intg(const intg& src):object(src),value(src.value){}
		/// Creates an intg with the indicated initial value.
		intg(int d):object(),value(d){}
		intg operator+(const intg& b) const { return intg(value+b.value); }
		intg operator+(int b) const { return intg(value+b); }
		intg operator-(const intg& b) const { return intg(value-b.value); }
		intg operator-(int b) const { return intg(value-b); }

		intg operator*(const intg& b) const { return intg(value*b.value); }
		intg operator*(int b) const { return intg(value*b); }

		intg operator/(const intg& b) const { return intg(value/b.value); }
		intg operator/(int b) const { return intg(value/b); }

		/// Mod operator
		intg operator%(const intg& b) const { return intg(value%b.value); }
		/// Mod operator
		intg operator%(int b) const { return intg(value%b); }

		bool operator==(const intg& b) const { return value == b.value; }
		bool operator==(int b) const { return value == b; }
		bool operator!=(const intg& b) const { return value != b.value; }
		bool operator!=(int b) const { return value != b; }

		bool operator>=(int b) const { return value >= b; }
		bool operator>=(const intg& b) const { return value >= b.value; }

		bool operator>(int b) const { return value > b; }
		bool operator>(const intg& b) const { return value > b.value; }

		bool operator<=(int b) const { return value <= b; }
		bool operator<=(const intg& b) const { return value <= b.value; }

		bool operator<(int b) const { return value < b; }
		bool operator<(const intg& b) const { return value < b.value; }

		const intg& operator=(const intg& b) { value = b.value; return *this; }
		const intg& operator=(int b) { value = b; return *this; }

		const intg& operator+=(const intg& b) { value += b.value; return *this; }
		const intg& operator+=(int b) { value += b; return *this; }

		const intg& operator-=(const intg& b) { value -= b.value; return *this; }
		const intg& operator-=(int b) { value -= b; return *this; }

		const intg& operator*=(const intg& b) { value *= b.value; return *this; }
		const intg& operator*=(int b) { value *= b; return *this; }

		const intg& operator/=(const intg& b) { value /= b.value; return *this; }
		const intg& operator/=(int b) { value /= b; return *this; }

		const intg& operator%=(const intg& b) { value %= b.value; return *this; }
		const intg& operator%=(int b) { value %= b; return *this; }

		const intg& operator++() { value++; return *this; }
		const intg& operator--() { value--; return *this; }
		const intg& operator++(int i) { value++; return *this; }
		const intg& operator--(int i) { value--; return *this; }

		object* clone() const { return new intg(*this); }
		~intg(){}
};

/**
A double type wrapped in an object.
*/
class dbl: public object 
{
	public:	
		double value;
		/// Creates a dbl with a value of 0.
		dbl():object(),value(0.0){}
		/// Copy constructor.
		dbl(const dbl& src):object(src),value(src.value){}
		/// Creates a dbl with the indicated initial value.
		dbl(double d):object(),value(d){}

		dbl operator+(const dbl& b) const { return dbl (value + b.value); }
		dbl operator+(double b) const { return dbl (value + b); }

		dbl operator-(const dbl& b) const { return dbl (value - b.value); }
		dbl operator-(double b) const { return dbl (value - b); }

		dbl operator*(const dbl& b) const { return dbl (value * b.value); }
		dbl operator*(double b) const { return dbl (value * b); }

		dbl operator/(const dbl& b) const { return dbl (value / b.value); }
		dbl operator/(double b) const { return dbl (value / b); }

		bool operator==(const dbl& b) const { return value == b.value; }
		bool operator==(double b) const { return value == b; }
		bool operator!=(const dbl& b) const { return value != b.value; }
		bool operator!=(double b) const { return value != b; }

		bool operator>=(double b) const { return value >= b; }
		bool operator>=(const dbl& b) const { return value >= b.value; }

		bool operator>(double b) const { return value > b; }
		bool operator>(const dbl& b) const { return value > b.value; }

		bool operator<=(double b) const { return value <= b; }
		bool operator<=(const dbl& b) const { return value <= b.value; }

		bool operator<(double b) const { return value < b; }
		bool operator<(const dbl& b) const { return value < b.value; }

		const dbl& operator=(const dbl& b) { value = b.value; return *this; }
		const dbl& operator=(double b) { value = b; return *this; }

		const dbl& operator+=(const dbl& b) { value += b.value; return *this; }
		const dbl& operator+=(double b) { value += b; return *this; }

		const dbl& operator-=(const dbl& b) { value -= b.value; return *this; }
		const dbl& operator-=(double b) { value -= b; return *this; }

		const dbl& operator*=(const dbl& b) { value *= b.value; return *this; }
		const dbl& operator*=(double b) { value *= b; return *this; }

		const dbl& operator/=(const dbl& b) { value /= b.value; return *this; }
		const dbl& operator/=(double b) { value /= b; return *this; }

		const dbl& operator++() { value++; return *this; }
		const dbl& operator--() { value--; return *this; }
		const dbl& operator++(int i) { value++; return *this; }
		const dbl& operator--(int i) { value--; return *this; }

		object* clone() const { return new dbl(*this); }

		~dbl(){}
}; 

/**
An STL string wrapped up as an object.
*/
class strng: public object
{
	public:
		/// The value of the string object.
		std::string value;
		/// Create a strng with a value of "".
		strng():
		object(),
		value("")
		{
		}
		/// Copy constructor.
		strng(const strng& src):
		object(src),
		value(src.value)
		{
		}
		/// Creates a strng with the indicated initial value.
		strng(const std::string& src):
		object(),
		value(src)
		{
		}
		/// Assignment operator.
		const strng& operator=(const strng& src)
		{
			value = src.value;
			return *this;
		}
		const strng& operator=(const std::string& src)
		{
			value = src;
			return *this;
		}
		/// Equality compares the string values.
		bool operator==(const std::string& y) const
		{
			return value == y;
		}
		bool operator==(const strng& y) const
		{
			return value == y.value;
		}
		object* clone() const 
		{ 
			return new strng(*this); 
		}
		~strng()
		{
		}
};

#endif
