/***************
Copyright (C) 2000, 2001, 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __adevs_vec_h_
#define __adevs_vec_h_
#include "adevs_config.h"

/**
An N dimensional, double valued vector.
*/
template <int N> class adevs_vec
{
	public:
		/// Creates a vector whose elements are 0.0.
		adevs_vec()
		{
			for (int i = 0; i < N; i++) u[i] = 0.0;
		}
		/// Copy constructor.
		adevs_vec(const adevs_vec<N>& src)
		{
			for (int i = 0; i < N; i++) u[i] = src.u[i];
		}
		/// Creates a vector whose elements are taken from the source array.
		adevs_vec(const double* src)
		{
			for (int i = 0; i < N; i++) u[i] = src[i];
		}
		/// Assign an array of N doubles to the elements of this vector.
		const adevs_vec<N>& operator=(const double* src)
		{
			for (int i = 0; i < N; i++) u[i] = src[i];
			return *this;
		}
		/// Assign a vector value to this vector.
		const adevs_vec<N>& operator=(const adevs_vec<N>& src)
		{
			for (int i = 0; i < N; i++) u[i] = src.u[i];
			return *this;
		}
		/// Return a reference to the kth element.
		double& operator[](int k) { return u[k]; }
		/// Get a read-only copy of the kth element.
		double get(int k) const { return u[k]; }
		/// Get a pointer to the array that represents the vector.
		const double* operator*() const { return u; }
		/// Returns the number of elements in the vector.
		int size() const { return N; }
		/// Element-wise addition.
		adevs_vec<N> operator+(const adevs_vec<N>& x) const
		{
			adevs_vec<N> v;
			for (int i = 0; i < N; i++) v[i] = u[i]+x.get(i); 
			return v;
		}
		/// Element-wise add and assign.
		const adevs_vec<N>& operator+=(const adevs_vec<N>& x)
		{
			for (int i = 0; i < N; i++) u[i] += x.get(i); 
			return *this;
		}
		/// Element-wise subtraction.
		adevs_vec<N> operator-(const adevs_vec<N>& x) const
		{
			adevs_vec<N> v;
			for (int i = 0; i < N; i++) v[i] = u[i]-x.get(i); 
			return v;
		}
		/// Element-wise subtract and assign.
		const adevs_vec<N>& operator-=(const adevs_vec<N>& x)
		{
			for (int i = 0; i < N; i++) u[i] -= x.get(i); 
			return *this;
		}
		/// Scalar multiplication.
		adevs_vec<N> operator*(double x) const
		{
			adevs_vec<N> v;
			for (int i = 0; i < N; i++) v[i] = u[i]*x;
			return v;
		}
		/// Scalar division.
		adevs_vec<N> operator/(double x) const
		{
			adevs_vec<N> v;
			for (int i = 0; i < N; i++) v[i] = u[i]/x;
			return v;
		}
		/// Computes the L1 norm.
		double norm1() const
		{
			double n = 0.0;
			for (int i = 0; i < N; i++)
				n += fabs(u[i]);
			return n;
		}
		/// Computes the L2 norm.
		double norm2() const
		{
			double n = 0.0;
			for (int i = 0; i < N; i++)
				n += u[i]*u[i];
			return sqrt(n);
		}
		/// Find the largest element in the vector.
		double max() const
		{
			double D = 0.0;
			for (int i = 0; i < N; i++)
				D = std::max(D,fabs(u[i]));
			return D;
		}
		/// Find the smallest element in the vector.
		double min() const
		{
			double D = DBL_MAX;
			for (int i = 0; i < N; i++)
				D = std::min(D,fabs(u[i]));
			return D;
		}
		/// Destructor.
		~adevs_vec(){}
	private:
		double u[N];
};

template <int N> adevs_vec<N> operator*(double x, const adevs_vec<N>& y)
{
	return y*x;
}

#endif
