/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2004 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __cellSpace_models_h_
#define __cellSpace_models_h_
#include "adevs_config.h"
#include "adevs_models.h"
#ifdef OK_TO_COMPILE_ADEVS_CELLSPACE

ADEVS_NAMESPACE_BEGIN

/**
Any model that will be added to a cellSpace should inherit this interface.  It provides
essential input and output ports and methods for establishing the location
of a cell within the cellspace.
*/
class cell_interface
{
	public:
		/**
		Input grid, where this cell is at 1,1,1.  To find the input from the
		cell displaced from this cell by +1 or -1, add the displacement value
		to 1 to get the correct input port index.  For example, the cell 
		at the relative position -1,-1,0 provides input on port 0,0,1.
		*/
		static const ADEVS_PORT_TYPE in [3][3][3];
		/// Output on this port will be sent to all neighboring cells.
		static const ADEVS_PORT_TYPE out;
		/// Directed output ports.  Indexing is as per the input ports.
		static const ADEVS_PORT_TYPE dout [3][3][3];
		/// Lookup table for finding offsets from port numbers.
		static const int offset_table[28][3];
		/**
		Compute a key value for a location.  Arguments should be such
		that 0 <= x < w and 0 <= y < h.
		*/
		static unsigned long get_key(int x, int y, int z, int w, int h)
		{
			return (unsigned long)(x+y*w+z*w*h);
		}
		/**
		Get the cell coordinates from a key.  The coordinate is assigned
		to the x, y, and z references passed to the method.
		*/
		static void get_coord(int& x, int& y, int& z, int w, int h, int key)
		{
			z = (int)ldiv(key,w*h).quot;
			unsigned long p = key-z*w*h;
			y = (int)ldiv(p,w).quot;
			x = p - y*w;
		}
};

/**
A 3D cellspace.  This class is meant to be useful for solving PDEs, simulating
next event cellspaces, and other types of models represented as spaces of discrete
interacting points.  It can support a six point (Von Neuman) or 26 point (Moore)
neighborhood.
*/
class cellSpace: public devsn
{
	public:

		/// A coordinate in the cell space.
		struct coord_t
		{
			/// Coordinate x, y, and z values.
			int x, y, z;
			/// Default constructor.
			coord_t(int x = 0, int y = 0, int z = 0):
			x(x),
			y(y),
			z(z)
			{
			}
			/// Assigne the values in a 3 element array to the coordinates x, y, and z.
			coord_t(const int* src):
			x(src[0]),
			y(src[1]),
			z(src[2])
			{
			}
			/// Copy constructor.
			coord_t(const coord_t& src):
			x(src.x),
			y(src.y),
			z(src.z)
			{
			}
			/// Assignment operator.
			const coord_t& operator=(const coord_t& src)
			{
				x = src.x;
				y = src.y;
				z = src.z;
				return *this;
			}
		};

		/// Flag indicating a six point neighborhood.
		static const int SIX_POINT;
		/// Flag indicating a twenty six point neighborhood.
		static const int TWENTY_SIX_POINT;
		/// Creates a 0x0x0 cellspace.
		cellSpace():
		devsn()
		{
			nx = new net_exec(this);
		}
		/// Copy constructor.
		cellSpace(const cellSpace& src):
		devsn(src)
		{
			nx = new net_exec(*(src.nx),this);
		}
		/// Assignment operator.
		const cellSpace& operator=(const cellSpace& src)
		{
			devsn::operator=(src);
			delete nx;
			nx = new net_exec(*(src.nx),this);
			return *this;
		}
		/// Create an W x H x D cellspace with NULL model entries in the cells.
		cellSpace(int type, int width, int height = 1, int depth = 1):
		devsn()
		{
			nx = new net_exec(this,width,height,depth,type);
		}
		/// Insert a model at the x,y,z position.
		void add(devs* model, int x, int y = 0, int z = 0) 
		{
			model->bsim_support.key =
				cell_interface::get_key(x,y,z,nx->getWidth(),nx->getHeight());
			nx->addModel(model,x,y,z); 
		}
		/// Get the network executive for this coupled model.
		netExec* getNetExec() { return nx; }
		/// Get the model at location x,y,z.
		const devs* getModel(int x, int y = 0, int z = 0) const
		{
			return nx->getModel(x,y,z);
		}
		/// Get a mutatable version of the model at x,y,z.
		devs* getMutableModel(int x, int y = 0, int z = 0)
		{
			return nx->getMutableModel(x,y,z);
		}
		/// Get the port number assigned to coordinate x, y, z.
		ADEVS_PORT_TYPE getPort(int x, int y = 0, int z = 0) const
		{
			return nx->getPort(x,y,z);
		}
		/// Get the coordinate assigned to a port.
		coord_t getCoordinates(ADEVS_PORT_TYPE port) const
		{
			return nx->getCoordinates(port);
		}
		/// Get the width of the cellspace.
		int getWidth() const { return nx->getWidth(); }
		/// Get the height of the cellspace.
		int getHeight() const { return nx->getHeight(); }
		/// Get the depth of the cellspace.
		int getDepth() const { return nx->getDepth(); }
		/// Destructor.
		~cellSpace(){ delete nx; }
		/// Returns a copy of this cellSpace.
		object* clone() const { return new cellSpace(*this); }
	private:	

		class net_exec: public netExec
		{
			public:
				net_exec(devsn* parent = NULL, 
				int width = 0, int height = 0, int depth = 0, int type = SIX_POINT);
				net_exec(const net_exec& src, devsn* parent);
				void init();
				void delta_int(){}
				void delta_ext(ADEVS_TIME_TYPE e, const adevs_bag<PortValue>& x){}
				void delta_conf(const adevs_bag<PortValue>& x){}
				void output_func(adevs_bag<PortValue>& y){}
				void gc_output(adevs_bag<PortValue>& g){}
				void gc_models(adevs_set<devs*>& g){}
				void getComponents(adevs_set<devs*>& c);
				void addModel(devs* d, int x, int y, int z);
				const devs* getModel(int x, int y, int z) const;
				devs* getMutableModel(int x, int y, int z);
				int getWidth() const { return w; }
				int getHeight() const { return h; }
				int getDepth() const { return d; }
				void route(const PortValue& pv, devs* model, adevs_bag<EventReceiver>& r);
				ADEVS_PORT_TYPE getPort(int x, int y, int z) const;
				ADEVS_PORT_TYPE getPort(const devs* model) const;
				coord_t getCoordinates(ADEVS_PORT_TYPE port) const;
				~net_exec();
			private:
				int w, h, d, type;
				devs**** space;

				void directed_internal_coupling(const PortValue& pv, devs* model, 
				adevs_bag<EventReceiver>& r);
				void undirected_internal_coupling(const PortValue& pv, devs* model, 
				adevs_bag<EventReceiver>& r);
				void external_input_coupling(const PortValue& pv, devs* model,
				adevs_bag<EventReceiver>& r);
				void external_output_coupling(const PortValue& pv, devs* model,
				adevs_bag<EventReceiver>& r);

				net_exec(const net_exec& src){}
				void operator=(const net_exec& src){}
		};
		net_exec* nx;
};

ADEVS_NAMESPACE_END

#endif
#endif
