/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001, 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __devssim_h_
#define __devssim_h_
#include "adevs_config.h"
#include "adevs_models.h"
#include "pdevs.h"

ADEVS_NAMESPACE_BEGIN

/**
The devssim functions as the root coordinator for a simulation.
In generally, it should not be sub-classed, except to override the
halt() method.
*/
class devssim 
{
	public:
		/**
		Create and initialize a simulator for the model with initial time t0.  
		*/
		devssim(devs* model, ADEVS_TIME_TYPE t0 = ADEVS_ZERO,
		ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON);
		/**
		Step the simulator by dt units of time.  Throws an adevs_exception
		if dt < 0 or if timeCurrent()+dt > timeNext().
		*/
		void step(ADEVS_TIME_TYPE dt);
		/**
		Steps the simulator after applying the PortValues in the 
		input list to the top level model.  The values of the PortValue
		objects are owned by the caller and can be freed when the step()
		method returns.  
		*/
		void step(ADEVS_TIME_TYPE dt, const adevs_bag<PortValue>& input);
		/**
		Execute the next simulation cycle and return.
		*/
		void execNextEvent();
		/**
		Run the simulation until the time of next event is greater than t_max.
		*/
		void run(ADEVS_TIME_TYPE t_max = ADEVS_INFINITY);
		/// Get the current simulation time.
		ADEVS_TIME_TYPE timeCurrent();
		/// Returns the time of next event.
		ADEVS_TIME_TYPE timeNext();
		/// Returns the time of the last event.
		ADEVS_TIME_TYPE timeLast();
		/// Returns true if tN = ADEVS_INFINITY, false otherwise.
		bool isPassive();
		/**
		Reset the simulator and reinitialize the model.
		*/
		virtual void reset(ADEVS_TIME_TYPE t0 = ADEVS_ZERO, 
		ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON);
		/**
		Returns true if the simulator should stop, false otherwise.  This method
		is called by the simulator at the end of each simulation
		cycle.  The default behavior is to return true when the model becomes
		passive.
		*/
		virtual bool halt();
		/**
		Destructor.  Does not delete the model.
		*/
		virtual ~devssim();
	private:
		devs* model;
		basic_sim* sim;
		ADEVS_TIME_TYPE t_current;

		devssim(const devssim& src){}
		void operator=(const devssim& src){}

};

ADEVS_NAMESPACE_END

#endif
