/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2005 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __par_pdevs_h__
#define __par_pdevs_h__
#include "pdevs.h"
#ifdef ADEVS_USE_POSIX_THREADS
#include <pthread.h>
#else
#include "empty_pthread.h"
#endif

ADEVS_NAMESPACE_BEGIN

/**
Basic computational element for the parallel simulation.  This class corresponds
to an LP in the standard PDES vocabulary.  It simulates a collection of DEVS models
using (ultimately) a risk-free optimistic approach.  The collection of simulation threads
is controlled, in a top down way, by the par_pdevs_coord class.  Note that only one
asynchronous method can run at any given time.  The wait() method will block until the
currently running asynchronous method completes.
*/
class sim_thread
{
	public:
		/**
		Creates the simulation thread and supporting data structures.  The thread lookup table
		should be an array of pointers to the sim_thread objects being used by the simulation
		engine.  The thread ID should be a unique number in [1,ADEVS_NUM_THRADS).
		*/
		sim_thread(int thrd_id, int num_thrds, ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON);
		/**
		Assign an atomic model to this thread initially.  The thread will place the model into its
		schedule and set the model's thread owner to point to this sim_thread object.  The model
		should only be modified by the thread which owns it.
		*/
		void initial_assignment(atomic* model);
		/**
		Activate a model that is already assigned to this thread.  Activating the model places it
		into the imminent set and assigns the supplied input.  
		This is done during event routing when input events originate from
		a model not local to the owning thread.
		*/
		void activate(atomic* model, PortValue& pv, int calling_thrd);
		/**
		Get the time of the next event for this thread.
		*/
		ADEVS_TIME_TYPE timeNext() const { return sched.minPriority(); }
		/**
		Compute the IO for the models assigned for this thread using the provided global virtual time.
		This method will run in a different thread than the caller.  The caller should use the wait()
		method to determine when the method has terminated.
		*/
		void computeInputOutput(ADEVS_TIME_TYPE gvt);
		/**
		Compute the next state for active models in this thread using the input values computed using
		the computeInputOutput() method.  The wait() method can be used to determine when the method
		call is completed.
		*/
		void computeDeltFunc();
		/**
		Do the end of cycle cleanup for this thread.  The wait() method can be used to determine when
		the method call is completed.
		*/
		void endOfCycle();
		/**
		Wait for the last asynchronous method call to complete.
		*/
		void wait();
		/**
		Destructor.  
		*/
		~sim_thread();

	private:
		/// Structure for storing information about cross-thread inputs
		struct model_io_t
		{
			atomic* model;
			PortValue pv;
		};
		/// The schedule for this thread
		adevs_sched sched;
		/// Set of models that this thread needs to compute state changes for
		adevs_bag<atomic*> imm;
		/// Bag of external inputs
		adevs_bag<model_io_t>* x_thrd_io;
		/// Last known global virtual time
		ADEVS_TIME_TYPE gvt;
		/// Time granule
		ADEVS_TIME_TYPE t_granule;
		/// Pool of input and output event bags
		object_pool<adevs_bag<PortValue> > io_pool;
		/// Pool of event routing bags
		object_pool<adevs_bag<EventReceiver> > recv_pool;
		/// Next method execution request 
		int run_method;
		/// Friend function runs the main thread loop
		friend void* sim_thread_main(void* arg);
		void dispatch(int method);
		void computeIO();
		void route(devsn* parent, devs* src, PortValue& pv);
		void computeNextState();
		void computeEndOfCycle();
		void activate_local(atomic* model, PortValue& pv);
		/// Thread handle
		pthread_t thr_hnd;
		int num_thrds;
		/// Conditional for starting/ending the execution of an asynchronous method
		pthread_cond_t run_cond;
		pthread_mutex_t cond_mtx;
		/// ID for this thread
		int thrd_id;
		/// No copy constructor
		sim_thread(const sim_thread&){}
		/// No assignment operator
		void operator=(const sim_thread&){}
};

/**
A parallelized multi-component coordinator.  This coordinator does not support the injection 
or extraction of external inputs, and it is not suitable for use in component based simulation
environments.
*/
class par_pdevs_coord: public basic_sim
{
	public:
		/**
		Construct a simulator for the supplied model.  The supplied time granule
		will be used for zero calculations (ADEVS_EPSILON by default).
		*/
		par_pdevs_coord(devsn* model, int num_thrds, ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON);
		/// Compute the model I/O for time t = tN.
		void computeInputOutput();
		/// This method is not support, and will throw an adevs_exception if it is called.
		const adevs_bag<PortValue>& getOutput();
		/**
		Compute the next state of this model for time t == tN().  Calls with any other time
		will fail with an adevs_exception.
		*/
		void computeDeltaFunc(ADEVS_TIME_TYPE t);
		/// Do garbage collection and rescheduling at the end of a simulation cycle.
		void endOfCycle();
		/**
		This method is not supported.  Calling it will result in an adevs_exception.
		*/
		void inject(const PortValue& x);
		/**
		This method is not supported.  Calling it will result in an adevs_exception.
		*/
		void inject (const adevs_bag<PortValue>& x);
		/// Get the simulator's time of next event.
		ADEVS_TIME_TYPE timeNext() const { return global_tN; }
		/// Get the simulator's time of last event.
		ADEVS_TIME_TYPE timeLast() const { return global_tL; }
		/// Destructor.  Does not destroy the model.
		virtual ~par_pdevs_coord();

	protected:
		/// Global virtual time
		ADEVS_TIME_TYPE global_tL;
		/// Time of last event
		ADEVS_TIME_TYPE global_tN;
		/// Array of simulation threads
		sim_thread** thrd;
		int num_thrds;
		/// Schedule the components of a network model
		void schedule(devsn* model);

		// No default constructor
		par_pdevs_coord(){}
		// No copy constructor
		par_pdevs_coord(const par_pdevs_coord&){}
		// No assignment operator
		void operator=(const par_pdevs_coord&){}
};

ADEVS_NAMESPACE_END

#endif
