/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2003 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __pdevs_h__
#define __pdevs_h__
#include "adevs_config.h"
#include "adevs_bag.h"
#include "adevs_set.h"
#include "adevs_sched.h"
#include "adevs_models.h"
#include "object_pool.h"

ADEVS_NAMESPACE_BEGIN

/**
Common simulator interface (based on the DEVS Core Interface specification).
*/
class basic_sim
{
	public:
		/**
		Constructor can be used to set the simulator time granule.  
		The default value is ADEVS_EPSILON.
		*/
		basic_sim(ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON):
		t_granule(t_granule)
		{
		}
		/// Get the time granule for this simulator.
		ADEVS_TIME_TYPE getTimeGranule() const { return t_granule; }
		/// Compute the model I/O for time t = tN().
		virtual void computeInputOutput() = 0;
		/**
		Get the output for this model previously computed with computeInputOutput().
		The output objects should be deleted during the next call to endOfCycle().
		*/
		virtual const adevs_bag<PortValue>& getOutput() = 0;
		/**
		Compute the next state of this model for time t <= tN().
		This method will apply all injected inputs at time t and use the results
		of the previous call to computeInputOutput() if t = tN().
		*/
		virtual void computeDeltaFunc(ADEVS_TIME_TYPE t) = 0;
		/// Do garbage collection at the end of a simulation cycle.
		virtual void endOfCycle() = 0;
		/**
		Inject an input for use during the next call to computeDeltaFunc()
		The input is owned by the caller and will not be deleted by the simulator.
		*/
		virtual void inject(const PortValue& x) = 0;
		/**
		Inject an input for use during the next call to computeDeltaFunc().
		The input is owned by the caller and will not be deleted by the simulator.
		*/
		virtual void inject (const adevs_bag<PortValue>& x) = 0;
		/// Get the simulator's time of next event.
		virtual ADEVS_TIME_TYPE timeNext() const = 0;
		/// Get the simulator's time of last event.
		virtual ADEVS_TIME_TYPE timeLast() const = 0;
		/// Destructor.  Does not deleted the model.
		virtual ~basic_sim()
		{
		}
	protected:
		ADEVS_TIME_TYPE t_granule;
		// No copy constructor.
		basic_sim(const basic_sim&){}
		// No assignment operator.
		void operator=(const basic_sim&){}
};

/**
A single component simulator.
*/
class pdevs_sim: public basic_sim
{
	public:
		/** 
		Construct a simulator for the supplied atomic model.
		*/
		pdevs_sim(atomic* model, ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON);
		/// Compute model output for time t = tN().
		void computeInputOutput();
		/// Get the model output computed via the last call to computeInputOutput().
		const adevs_bag<PortValue>& getOutput() { return y; }
		/**
		Compute next state of the model at time t using input supplied 
		via the inject method.
		*/
		void computeDeltaFunc(ADEVS_TIME_TYPE t);
		/// End of cycle garbage collection.
		void endOfCycle();
		/**
		Inject an input for use during the next call to computeDeltaFunc().
		The input is owned by the caller and will not be deleted by the simulator.
		*/
		void inject(const PortValue& x) { this->x.add(x); }
		/**
		Inject an input for use during the next call to computeDeltaFunc().
		The input is owned by the caller and will not be deleted by the simulator.
		*/
		void inject (const adevs_bag<PortValue>& x) { this->x.append(x); }
		/// Get the simulator's time of next event.
		ADEVS_TIME_TYPE timeNext() const { return model->timeNext(); }
		/// Get the time of last event.
		ADEVS_TIME_TYPE timeLast() const { return model->timeLast(); } 
		/// Destructor.
		~pdevs_sim();

	protected:
		atomic* model;
		adevs_bag<PortValue> x;
		adevs_bag<PortValue> y;

		// No default constructor.
		pdevs_sim(){}
		// No copy constructor.
		pdevs_sim(const pdevs_sim&){}
		// No assignment operator.
		void operator=(const pdevs_sim&){}
};

/**
A multi-component coordinator.
*/
class pdevs_coord: public basic_sim
{
	public:
		/**
		Construct a simulator for the supplied model.  The supplied time granule
		will be used for zero calculations (ADEVS_EPSILON by default).
		*/
		pdevs_coord(devsn* model, ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON);
		/// Compute the model I/O for time t = tN().
		void computeInputOutput();
		/// Get the output for this model previously computed with computeInputOutput().
		const adevs_bag<PortValue>& getOutput() { return top_model_output; }
		/**
		Compute the next state of this model for time t <= tN().
		Will apply all injected inputs at time t and use the results
		of the previous call to computeInputOutput() if t = tN().
		*/
		void computeDeltaFunc(ADEVS_TIME_TYPE t);
		/// Do garbage collection and rescheduling at the end of a simulation cycle.
		void endOfCycle();
		/**
		Inject an input for use during the next call to computeDeltaFunc().
		The input is owned by the caller and will not be deleted by the simulator.
		*/
		void inject(const PortValue& x) { top_model_input.add(x); }
		/**
		Inject an input for use during the next call to computeDeltaFunc().
		The input is owned by the caller and will not be deleted by the simulator.
		*/
		void inject (const adevs_bag<PortValue>& x) { top_model_input.append(x); }
		/// Get the simulator's time of next event.
		ADEVS_TIME_TYPE timeNext() const { return global_tN; }
		/// Get the simulator's time of last event.
		ADEVS_TIME_TYPE timeLast() const { return global_tL; } 
		/// Destructor.  Does not destroy the model.
		virtual ~pdevs_coord();

	protected:
		devsn* top_model;
		ADEVS_TIME_TYPE global_tL;
		ADEVS_TIME_TYPE global_tN;
		/// The atomic model's active flag is used to determine membership in the bag
		adevs_bag<atomic*> imm;
		/// List of network executives that appear in the imminent set.
		adevs_bag<netExec*> nx_imm;
		/// The event schedule
		adevs_sched sched;
		/// I/O for the top level model
		adevs_bag<PortValue> top_model_input;
		adevs_bag<PortValue> top_model_output;
		/// Pools of preallocated, commonly used objects
		object_pool<adevs_bag<PortValue> > io_pool;
		object_pool<adevs_bag<EventReceiver> > recv_pool;

		/// Schedule the components of a network model
		void schedule(devsn* model);
		/// Route an event generated by the source model contained in the parent model
		void route(devsn* parent, devs* src, PortValue& pv);
		/// Reschedule an atomic component
		void reschedule_atomic(atomic* model, bool force_passive = false);
		/// Compute a state change for an atomic model
		void compute_atomic_delta_func(atomic* model, ADEVS_TIME_TYPE t);
		/**
		Apply structure changes.  State changes for the network executives are computed
		as part of applying the structural change.  The component_set_changed
		flag should be true if any of the models appearing in the nx_imm set
		have their componentSetChanged() flags set to true.
		*/
		void compute_structure_changes(bool component_set_changed, ADEVS_TIME_TYPE t);
		/// Recursively remove a model from the simulation data structures.
		void remove_model(devs* model);

		// No default constructor
		pdevs_coord(){}
		// No copy constructor
		pdevs_coord(const pdevs_coord&){}
		// No assignment operator
		void operator=(const pdevs_coord&){}
};

ADEVS_NAMESPACE_END

#endif
