/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2005 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __pdevssim_h_
#define __pdevssim_h_
#include "adevs_config.h"
#include "adevs_models.h"
#include "par_pdevs.h"

ADEVS_NAMESPACE_BEGIN

/**
<P>The pdevssim functions can be used to simulate large multi-component models on small
parallel computers.  This parallel simulator will execute simultaneous events in 
parallel.  This means that state transition functions are executed in parallel, output
functions are executed in parallel, garbage collection is done in parallel, and
event routing is done in parallel.</P>
<P>There are four constraints that parallel execution places on a model.  These are
<li>It must be safe to execute model state transition in parallel and output functions in parallel.
If you designed your model's properly, this isn't a problem (because models don't share
data except through input and output ports).  However, in the rare instance that you
have some unsafe sharing going on, it needs to be fixed before the parallel simulator is used.
<li>The event routing method for your network model must be reentrant.  Because events are routed in parallel, the event routing method can be called simutaneously by multiple threads.  The coupled models that come with adevs are safe for use with the parallel simulator.  Custom models will need to make sure that the event routing method is thread safe (ideally, by being reentrant).
<li>Garbage collection methods must be able to execute in parallel.
<li>Your model is not a dynamic structure model.  There is no support for parallel simulation
of dynamic structure models.
</P>
*/
class pdevssim 
{
	public:
		/**
		Create and initialize a simulator for the model with initial time t0.  
		The model will be simulated using the specified number of threads.
		If NUMA support was enabled, then models will be simulated on their
		assigned nodes.  Otherwise, the assignment of models to threads will be
		as desired, but there is no way of controlling how threads are assigned
		to processes. 
		Only coupled models can be executed in paralle, because there is no 
		parallel potential in an atomic model
		(at least, not using the parallel algorithm implemented here).
		Dynamic structure models are NOT supported by the parallel simulation engine.
		*/
		pdevssim(devsn* model, int num_threads, ADEVS_TIME_TYPE t0 = ADEVS_ZERO,
		ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON);
		/**
		Run the simulation until the time of next event is greater than t_max.
		*/
		void run(ADEVS_TIME_TYPE t_max = ADEVS_INFINITY);
		/**
		Run the simulation until the next synchronization point.
		*/
		void step();
		/**
		Return GVT at the current synchronization point.
		*/
		ADEVS_TIME_TYPE timeNext();
		/**
		Reset the simulator and reinitialize the model.
		*/
		virtual void reset(int num_threads, ADEVS_TIME_TYPE t0 = ADEVS_ZERO, 
		ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON);
		/**
		Destructor.  Does not delete the model.
		*/
		virtual ~pdevssim();
	private:
		devsn* model;
		par_pdevs_coord* sim;

		pdevssim(const pdevssim& src){}
		void operator=(const pdevssim& src){}

};

ADEVS_NAMESPACE_END

#endif
