/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999-2003 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __prefab_models_h_
#define __prefab_models_h_
#include "adevs_config.h"
#include "adevs_models.h"

ADEVS_NAMESPACE_BEGIN

/**
A generic coupled model that can be constructed using
any netExec derived NET_EXEC class whose constructor
accepts a single argument which is the parent network
model.
*/
template <class NET_EXEC> class coupled: public devsn
{
	public:
		coupled():devsn(),nx(this){}
		netExec* getNetExec() { return &nx; }
		~coupled(){}
	private:
		NET_EXEC nx;
};

/**
Create a network model that will use any network executive
that is provided to it.  The network executive is
adopted by the network model and will be destroyed when
the network model is destroyed.
*/
class network_model: public devsn
{
	public:
		network_model(netExec* exec):
		devsn(),
		net_exec(exec)
		{
		}
		netExec* getNetExec()
		{
			return net_exec;
		}
		~network_model()
		{
			delete net_exec;
		}
	private:
		netExec* net_exec;
};

/**
The staticDigraph class can be used to create
static structure digraph models.  The copy constructor,
clone method, and assignment operator rely on the
component models supporting the clone() method.
*/
class staticDigraph: public devsn 
{
	public:
		staticDigraph():
		devsn()
		{
			nx = new net_exec(this);
		}
		staticDigraph(const staticDigraph& src):
		devsn(src)
		{
			nx = new net_exec(*(src.nx),this);
		}
		const staticDigraph& operator=(const staticDigraph& src)
		{
			devsn::operator=(src);
			delete nx;
			nx = new net_exec(*(src.nx),this);
			return *this;
		}
		/// Add a model to the network
		void add(devs* model) 
		{ 
			nx->addModel(model); 
		}
		/// Couple the src model to the destination model
		void couple(devs* src, ADEVS_PORT_TYPE srcPort, devs* dst, ADEVS_PORT_TYPE dstPort)
		{
			nx->addCoupling(src,srcPort,dst,dstPort);
		}
		netExec* getNetExec() { return nx; }
		~staticDigraph(){ delete nx; }
		object* clone() const { return new staticDigraph(*this); }

	private:	
		class net_exec: public netExec
		{
			public:
				net_exec(devsn* parent = NULL):netExec(parent){}
				net_exec(const net_exec& src, devsn* parent);
				void init();
				void delta_int(){}
				void delta_ext(ADEVS_TIME_TYPE e, const adevs_bag<PortValue>& x){}
				void delta_conf(const adevs_bag<PortValue>& x){}
				void output_func(adevs_bag<PortValue>& y){}
				void gc_output(adevs_bag<PortValue>& g){}
				void gc_models(adevs_set<devs*>& g){}
				void getComponents(adevs_set<devs*>& c);
				void addModel(devs* d);
				void addCoupling(devs* src, ADEVS_PORT_TYPE srcPort, 
					devs* dst, ADEVS_PORT_TYPE dstPort);
				void route(const PortValue& pv, devs* model, adevs_bag<EventReceiver>& r);
				~net_exec();
			private:
				adevs_set<devs*> models;
				adevs_hashtable<EventReceiver,adevs_bag<EventReceiver>,EventReceiverHashFunc >
					graph;
				net_exec(const net_exec& src){}
				void operator=(const net_exec& src){}
		};
		net_exec* nx;
		
	friend class net_exec;
};

ADEVS_NAMESPACE_END

#endif
