/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __qode_h_
#define __qode_h_
#include "adevs_config.h"

ADEVS_NAMESPACE_BEGIN

/**
A first order quantized ODE solver based on explicit Euler.  
*/
class qode
{
	public:
		/// Create a solver for an ODE.
		qode():
		q(0.0),
		dq(0.0),
		D(0.0),
		ql(0.0),
		N(0)
		{
			setParams(1.0);
		}
		/// Copy constructor
		qode(const qode& src):
		q(src.q),
		dq(src.dq),
		D(src.D),
		ql(src.ql),
		N(src.N)
		{}
		/// Assigment operator
		const qode& operator=(const qode& src)
		{
			q = src.q;
			dq = src.dq;
			D = src.D;
			ql = src.ql;
			N = src.N;
			return *this;
		}
		/**
		Initial the integrator set.  q0 is the initial state, dq0 is 
		the initial derivative.
		*/
		void init(const double q0, const double dq0)
		{
			N = 0;
			ql = q = q0;
			dq = dq0;
		}
		/**
		Integrate over dt (dt <= ta()) units of time with derivatives dq.  
		*/
		void integ(double dt, const double dq_)
		{
			// Get time to next boundary crossing
			double sigma = compute_ta();
			// update state
			if (dt >= sigma)
			{
				if (dq > 0.0) N++;
				else if (dq < 0.0) N--;
				q = ql + (double)N*D;
			}
			else
			{
				q += dq*dt;
			}
			dq = dq_;
		}
		/**
		Integrate over dt (dt <= ta()) units of time with unchanged derivatives.  
		*/
		void integ(double dt)
		{
			integ(dt,dq);
		}
		/**
		Return the quantized output at time tL + ta(),
		where tL was the time of the last state change.
		*/
		double out() const
		{
			if (dq > 0.0) return ql + (double)(N+1)*D;
			else if (dq < 0.0) return ql + (double)(N-1)*D;
			else return ql + (double)N*D;
		}
		/// Return the quantized output at time tL + dt, dt <= ta()
		double out(double dt) const
		{
			if (dt < compute_ta()) return ql + (double)N*D;
			else return out();
		}
		/// Return the state of the integrator at time tL
		double state() const
		{
			return q;
		}
		/// Get the quantized state of the integrator
		double qstate() const
		{
			return ql + (double)N*D;
		}
		/// The time advance function for the ODE. 
		double ta() const
		{
			return compute_ta();
		}
		/// Set the integration quantum
		void setParams(double D_)
		{
			D = D_;
		}
		~qode(){}
	private:
		// State variable 
		double q;
		// Input variable (derivative) 
		double dq;
		// Quantum size
		double D;
		// Initial state
		double ql;
		// Last output level
		int N;

		double compute_ta() const
		{
			double qq = q - (ql+(double)N*D);
			double sigma = ADEVS_INFINITY;
			if (fabs(dq) > ADEVS_EPSILON)
			{
				// time to reach lower level (q - (-d))
				if (dq < 0.0)
					sigma = fabs((qq+D)/dq);
				// time to reach upper level (d - q)
				else if (dq > 0.0)
					sigma = fabs((D-qq)/dq);
				// otherwise, no motion so passivate
			}
			if (sigma <= ADEVS_EPSILON) sigma = 0.0;
			return sigma;
		}
};

ADEVS_NAMESPACE_END

#endif
