/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __qoden_h_
#define __qoden_h_
#include "adevs_config.h"
#include "adevs_vec.h"

ADEVS_NAMESPACE_BEGIN

/**
A first order quantized ODE solver based on explicit Euler.  
The qoden uses a set of quantized integrators
to provide a solution to a set of ordinary differential equations.
This solver differs from the qode solver in that it uses a box in
N-dimensional space to determine significant events.  
*/
template <int N> class qoden
{
	public:
		/// Create a solver for a set of N ODEs.
		qoden():
		sigma(ADEVS_INFINITY)
		{
			setParams(1.0);
		}
		/// Copy constructor
		qoden(const qoden<N>& src):
		q(src.q),
		dq(src.dq),
		sigma(src.sigma),
		D(src.D),
		ql(src.ql)
		{}
		/// Assigment operator
		const qoden<N>& operator=(const qoden<N>& src)
		{
			q = src.q;
			dq = src.dq;
			sigma = src.sigma;
			D = src.D;
			ql = src.ql;
			return *this;
		}
		/**
		Initial the integrator set.  q0 is the initial state, dq0 are
		the initial derivatives.
		*/
		void init(const double* q0, const double* dq0)
		{
			for (int i = 0; i < N; i++) init(i,q0[i],dq0[i]);
		}
		/// Initialize the jth integrator
		void init(unsigned j, double q0, double dq0 = 0.0)
		{
			ql[j] = q[j] = q0;
			dq[j] = dq0;
			compute_ta();
		}
		/**
		Integrate over dt (dt <= ta()) units of time with derivatives dq.  
		*/
		void integ(double dt, const double* dq_)
		{
			// update state
			q += dq*dt;
			dq = dq_;
			// if reached quantum level, set last quantum point to current point
			if (dt >= sigma)
				ql = q;
			// compute time advance
			compute_ta();
		}
		/**
		Integrate over dt (dt <= ta()) units of time with unchanged derivatives.  
		*/
		void integ(double dt)
		{
			integ(dt,*dq);
		}
		/**
		Return the quantized output for integrator j at time tL + ta(),
		where tL was the time of the last state change.
		*/
		double out(unsigned j = 0) const
		{
			return (q.get(j)+dq.get(j)*sigma);
		}
		/// Return the quantized output for integrator j at time tL + dt, dt <= ta()
		double out(unsigned j, double dt) const
		{
			if (dt < sigma) return q.get(j);
			else return out(j);
		}
		/// Return the state of the integrator at time tL
		double state(unsigned j = 0) const
		{
			return q.get(j);
		}
		/// Get the quantized state of the integrator
		double qstate(unsigned j = 0) const
		{
			return ql.get(j);
		}
		/// The time advance function for the ODE set. 
		double ta() const
		{
			return sigma;
		}
		/// Set the parameters for iterator j (see qintg)
		void setParams(unsigned j, double D_)
		{
			D[j] = D_;
		}
		/// Set the parameters for all integrators (see qintg)
		void setParams(double D_)
		{
			for (int i = 0; i < N; i++) setParams(i,D_);
		}
		~qoden(){}
	private:
		// State variable set
		adevs_vec<N> q;
		// Input variable (derivative) set
		adevs_vec<N> dq;
		// Time to next boundary crossing
		double sigma;
		// Quantum limit in each dimension
		adevs_vec<N> D;
		// Last output position
		adevs_vec<N> ql;

		void compute_ta()
		{
			adevs_vec<N> qq(q-ql);
			// time for line of travel to reach each edge of quantum block
			// take min. of all to get time of next event
			sigma = ADEVS_INFINITY;
			for (int i = 0; i < N; i++)
			{
				if (fabs(dq[i]) > ADEVS_EPSILON)
				{
					// time to reach lower level of box (q - (-d))
					if (dq[i] < 0.0)
						sigma = std::min(sigma,fabs((qq[i]+D[i])/dq[i]));
					// time to reach upper level of the box (d - q)
					else if (dq[i] > 0.0)
						sigma = std::min(sigma,fabs((D[i]-qq[i])/dq[i]));
				}
				// otherwise, no motion so passivate
			}
			if (sigma <= ADEVS_EPSILON) sigma = 0.0;
		}
};

ADEVS_NAMESPACE_END

#endif
