/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __xdevssim_h_
#define __xdevssim_h_
#include "adevs_config.h"
#include "adevs_models.h"
#include "pdevs.h"
ADEVS_NAMESPACE_BEGIN

/**
This simulator class is meant to be used as a component
within another simulation environment (for example, HLA or DIS). 
It implements a reduced version of the coreSimulatorInterface
from the scalable DEVS interface.  The simplification
made here involves removing timestamps from all calls
except deltFunc(t).  It is assumed that injected inputs
and I/O computations occur at the next deltFunc() and
nextTN() respectively.
*/
class xdevssim 
{
	public:
		/** 
		Construct a simulator for the supplied model.  
		*/
		xdevssim(devs* model):
		IO_up_to_date(false),
		sim(NULL),
		model(model)
		{
		}
		/**
		Initialize the simulator prior to simulation start.  This will
		cause the init() method of every component model to be executed.
		The value t0 is the initial simulation time.  The time granule
		is the difference between event time stamps that is treated as
		being effectively equal to zero.
		*/
		virtual void initialize(ADEVS_TIME_TYPE t0 = ADEVS_ZERO,
		ADEVS_TIME_TYPE t_granule = ADEVS_EPSILON);
		/**
		Finalize the simulator at the end of a run.  The simulation engine
		will call no other methods after finalize has been called.
		*/
		virtual void finalize(){}
		/**
		Inject an input to be applied at then next deltFunc(t) call.
		The input is either a single PortValue pair or a bag of PortValue
		pairs.  The input value is owned by the caller and can be freed after
		the next call to deltFunc().
		*/
		void inject(const PortValue& x) { input.add(x); }
		/**
		Same as inject, but injects an entire bag of events.  The event values
		are owned by the caller and can be freed after the next call to
		deltFunc().
		*/
		void inject(const adevs_bag<PortValue>& x) { input.append(x); }
		/// Get the time of next event (tL + ta()).
		ADEVS_TIME_TYPE nextTN() const { return sim->timeNext(); }
		/**
		Compute the top-level model output at time nextTN().  
		*/
		const adevs_bag<PortValue>& computeInputOutput();
		/**
		Advance the simulation clock and compute the model state at time t <= nextTN().
		If t > nextTN(), then the simulator will throw an adevs_exception.
		*/
		void deltFunc(ADEVS_TIME_TYPE t);
		/// Get the model attached to this simulator.
		const devs* getModel() const { return model; }
		/// Destructor.  Does not destroy the model.
		virtual ~xdevssim();

	private:
		bool IO_up_to_date;
		basic_sim* sim;
		devs* model;
		adevs_bag<PortValue> input;

		xdevssim(const xdevssim& src){}
		void operator=(const xdevssim& src){}

};

ADEVS_NAMESPACE_END

#endif
