/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2005 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "adevs_sched.h"

ADEVS_NAMESPACE::adevs_sched::adevs_sched(int capacity):
capacity(capacity),
size(0)
{
	heap = new heap_element[capacity];
}

void ADEVS_NAMESPACE::adevs_sched::schedule(atomic* model, ADEVS_TIME_TYPE priority)
{
	/**
	Check for existance in the queue.  This check will catch an uninitialized
	q_index, but if it was mucked with outside of this class, then we will have
	problems.
	*/
	bool enqueued = false;
	if (model->asim_support.q_index > 0 && model->asim_support.q_index < capacity)
	{
		enqueued = (heap[model->asim_support.q_index].item == model);
	}
	/**
	If it is being added, then increment the size and check for
	capacity overflow.  Once this is done, actually add the model to the
	schedule.
	*/
	if (!enqueued && priority < ADEVS_INFINITY)
	{
		size++;
		if (size == capacity)
		{
			enlarge();
		}
		int to_move = percolate_up(size,priority);
		heap[to_move].priority = priority;
		heap[to_move].item = model;
		model->asim_support.q_index = to_move;
	}
	/**
	If the model is enqueued and just needs to be reschedule
	*/
	else if (enqueued && priority < ADEVS_INFINITY)
	{
		int to_move = model->asim_support.q_index;
		if (heap[to_move].priority > priority)
		{
			to_move = percolate_up(to_move,priority);
		}
		else if (heap[to_move].priority < priority)
		{
			to_move = percolate_down(to_move,priority);
		}
		else
		{
			// Don't do anything if the priority hasn't changed
			return;
		}
		heap[to_move].priority = priority;
		heap[to_move].item = model;
		model->asim_support.q_index = to_move;
	}
	/**
	Otherwise, if the model is enqueued and its time of next event is infinite,
	then it needs to be removed.
	*/
	else if (enqueued)
	{
		int to_move = model->asim_support.q_index;
		ADEVS_TIME_TYPE min_priority = minPriority();
		to_move = percolate_up(to_move,min_priority);
		heap[to_move].priority = min_priority;
		heap[to_move].item = model;
		model->asim_support.q_index = to_move;
		removeMinimum();
	}
	/**
	Otherwise, the model is not enqueued and its passive, so don't do
	anything.
	*/
}

void ADEVS_NAMESPACE::adevs_sched::removeMinimum()
{
	// Don't do anything if the heap is already empty
	if (size == 0)
	{
		return;
	}
	size--;
	// Set index to -1 to indicate that this model is no longer in the schedule
	heap[1].item->asim_support.q_index = -1;
	heap[1].item = NULL;
	heap[1].priority = ADEVS_INFINITY;
	if (size > 0)
	{
		heap[1] = heap[size+1];
		int i = percolate_down(1,heap[size+1].priority);
		heap[i].item = heap[size+1].item;
		heap[i].priority = heap[size+1].priority;
		heap[i].item->asim_support.q_index = i;
		heap[size+1].item = NULL;
	}
}

void ADEVS_NAMESPACE::adevs_sched::clear()
{
	for (int i = 0; i < size; i++)
	{
		heap[i+1].item = NULL;
		heap[i+1].priority = ADEVS_INFINITY;
	}
	size = 0;
}

ADEVS_NAMESPACE::adevs_sched::~adevs_sched()
{
	delete [] heap;
}

void ADEVS_NAMESPACE::adevs_sched::enlarge()
{
	heap_element* rheap = new heap_element[capacity*2];
	for (int i = 0; i < capacity; i++)
	{
		rheap[i] = heap[i];
	}
	capacity *= 2;
	delete [] heap;
	heap = rheap;
}

int ADEVS_NAMESPACE::adevs_sched::percolate_down(int index, ADEVS_TIME_TYPE priority)
{
	int child;
	int i = index;
	for (; i*2 <= size; i = child) 
	{
		child = i*2;
		if (child != size && heap[child+1].priority < heap[child].priority) 
		{
			child++;
		}
		if (priority > heap[child].priority) 
		{
			heap[i] = heap[child];
			heap[i].item->asim_support.q_index = i;
		}
		else break;
	}
	return i;
}

int ADEVS_NAMESPACE::adevs_sched::percolate_up(int index, ADEVS_TIME_TYPE priority)
{
	int i = index;
	while (i/2 != 0 && heap[i/2].priority >= priority) 
	{
		heap[i] = heap[i/2];
		heap[i].item->asim_support.q_index = i;
		i = i/2;
	}
	return i;
}
