/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2003 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "devssim.h"
USING_NAMESPACE_ADEVS

ADEVS_NAMESPACE::devssim::devssim(devs* model, ADEVS_TIME_TYPE t0, 
ADEVS_TIME_TYPE t_granule):
model(model),
sim(NULL)
{
	reset(t0,t_granule);
}

void ADEVS_NAMESPACE::devssim::execNextEvent()
{
	if (sim->timeNext() < ADEVS_INFINITY)
	{
		sim->computeInputOutput();
		sim->computeDeltaFunc(sim->timeNext());
		sim->endOfCycle();
	}
	t_current = sim->timeLast();
}
	
void ADEVS_NAMESPACE::devssim::step(ADEVS_TIME_TYPE dt)
{
	if (dt < ADEVS_ZERO) THROW_BASE_EXCEPTION("dt < 0");
	ADEVS_TIME_TYPE t_max = t_current+dt;
	while (sim->timeNext() <= t_max)
	{
		sim->computeInputOutput();
		sim->computeDeltaFunc(sim->timeNext());
		sim->endOfCycle();
	}
	t_current = t_max;
}

void ADEVS_NAMESPACE::devssim::step(ADEVS_TIME_TYPE dt, const adevs_bag<PortValue>& input)
{
	if (dt < ADEVS_ZERO) THROW_BASE_EXCEPTION("dt < 0");
	ADEVS_TIME_TYPE t_max = t_current+dt;
	while (sim->timeNext() < t_max)
	{
		sim->computeInputOutput();
		sim->computeDeltaFunc(sim->timeNext());
		sim->endOfCycle();
	}
	sim->inject(input);
	if (t_max - timeNext() <= sim->getTimeGranule())
	{
		sim->computeInputOutput();
	}
	sim->computeDeltaFunc(t_max);
	sim->endOfCycle();
	t_current = t_max;
}

void ADEVS_NAMESPACE::devssim::run(ADEVS_TIME_TYPE t_max)
{
	while (sim->timeNext() <= t_max && !halt())
	{
		sim->computeInputOutput();
		sim->computeDeltaFunc(sim->timeNext());
		sim->endOfCycle();
	}
	t_current = t_max;
}

ADEVS_TIME_TYPE ADEVS_NAMESPACE::devssim::timeCurrent()
{
	return t_current;
}

ADEVS_TIME_TYPE ADEVS_NAMESPACE::devssim::timeNext()
{
	return sim->timeNext();
}

ADEVS_TIME_TYPE ADEVS_NAMESPACE::devssim::timeLast()
{
	return sim->timeLast();
}

void ADEVS_NAMESPACE::devssim::reset(ADEVS_TIME_TYPE t0, ADEVS_TIME_TYPE t_granule)
{
	if (sim != NULL) delete sim;
	atomic* amodel = model->typeIsAtomic();
	if (amodel != NULL)
	{
		amodel->asim_support.tL = t0;
		amodel->asim_support.t = t0;
		amodel->initialize(t0);
		sim = new pdevs_sim(amodel,t_granule);
	}
	else
	{
		model->initialize(t0);
		sim = new pdevs_coord(model->typeIsNetwork(),t_granule);
	}
	t_current = t0;
}

bool ADEVS_NAMESPACE::devssim::isPassive()
{
	return sim->timeNext() >= ADEVS_INFINITY;
}

bool ADEVS_NAMESPACE::devssim::halt()
{
	return sim->timeNext() >= ADEVS_INFINITY;
}

ADEVS_NAMESPACE::devssim::~devssim()
{
	if (sim != NULL)
	{
		delete sim;
	}
}
