/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2005 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "par_pdevs.h"
#include "adevs_linux_numa_wrapper.h"
USING_NAMESPACE_ADEVS

ADEVS_NAMESPACE::par_pdevs_coord::par_pdevs_coord(devsn* model, int num_thrds, ADEVS_TIME_TYPE t_granule):
basic_sim(t_granule),
num_thrds(num_thrds)
{
	// Indexing variable
	int i;

	thrd = new sim_thread*[num_thrds];
	for (i = 0; i < num_thrds; i++)
	{
		// Make sure new thread data is allocated to the proper memory blocks and
		// runs on the correct node.
		adevs_linux_numa_set_preferred(i);
		adevs_linux_numa_run_on_node(i);
		// Create and start the thread
		thrd[i] = new sim_thread(i,num_thrds,t_granule);
	}
	// Go ahead with local data initialization
	adevs_linux_numa_set_preferred(-1);
	adevs_linux_numa_run_on_node(-1);
	global_tL = ADEVS_ZERO;
	global_tN = ADEVS_INFINITY;
	schedule(model);
	for (i = 0; i < num_thrds; i++)
	{
		if (thrd[i]->timeNext() < global_tN)
		{
			global_tN = thrd[i]->timeNext();
		}
	}
}

void ADEVS_NAMESPACE::par_pdevs_coord::schedule(devsn* model)
{
	// Schedule the network executive.  This will cause an exception later on
	// if the network executive actually becomes active.
	thrd[rand()%num_thrds]->initial_assignment(model->getNetExec()->typeIsAtomic());
	// Update the global time of last event
	if (model->getNetExec()->timeLast() > global_tL)
	{
		global_tL = model->getNetExec()->timeLast();
	}
	// Schedule each component model
	adevs_set<devs*> components;
	model->getNetExec()->getComponents(components);
	for (int i = 0; i < components.getSize(); i++)
	{
		// Atomic models are placed into the schedule
		if (components[i]->typeIsAtomic() != NULL)
		{
			// Assign models to their preferred threads, or randomly if no
			// placement is specified
			int thrd_to_assign = 
				components[i]->typeIsAtomic()->asim_support.preferred_thread;
			if (thrd_to_assign < 0 || thrd_to_assign >= num_thrds)
			{
				thrd_to_assign = rand()%num_thrds;
			}
			thrd[thrd_to_assign]->initial_assignment(components[i]->typeIsAtomic());
		}
		// Network models are scheduled recursively
		else
		{
			schedule(components[i]->typeIsNetwork());
		}
	}
}


void ADEVS_NAMESPACE::par_pdevs_coord::computeInputOutput()
{
	int i;
	for (i = 0; i < num_thrds; i++)
	{
		thrd[i]->computeInputOutput(global_tN);
	}
	for (i = 0; i < num_thrds; i++)
	{
		thrd[i]->wait();
	}
}

const adevs_bag<PortValue>& ADEVS_NAMESPACE::par_pdevs_coord::getOutput()
{
	THROW_BASE_EXCEPTION("par_devs_coord::getOutput() is not supported")
}

void ADEVS_NAMESPACE::par_pdevs_coord::computeDeltaFunc(ADEVS_TIME_TYPE t)
{
	if (t != global_tN)
	{
		THROW_BASE_EXCEPTION("t != tN in  par_devs_coord::computeDeltaFunc()")
	}
	int i;
	for (i = 0; i < num_thrds; i++)
	{
		thrd[i]->computeDeltFunc();
	}
	for (i = 0; i < num_thrds; i++)
	{
		thrd[i]->wait();
	}
}

void ADEVS_NAMESPACE::par_pdevs_coord::endOfCycle()
{
	global_tL = global_tN;
	global_tN = ADEVS_INFINITY;
	int i;
	for (i = 0; i < num_thrds; i++)
	{
		thrd[i]->endOfCycle();
	}
	for (i = 0; i < num_thrds; i++)
	{
		thrd[i]->wait();
		if (thrd[i]->timeNext() < global_tN)
		{
			global_tN = thrd[i]->timeNext();
		}
	}
}

void ADEVS_NAMESPACE::par_pdevs_coord::inject(const PortValue& x)
{
	THROW_BASE_EXCEPTION("par_devs_coord::inject(const PortValue&) is not supported")
}

void ADEVS_NAMESPACE::par_pdevs_coord::inject (const adevs_bag<PortValue>& x)
{
	THROW_BASE_EXCEPTION("par_devs_coord::inject(const adevs_bag<PortValue>&) is not supported")
}

ADEVS_NAMESPACE::par_pdevs_coord::~par_pdevs_coord()
{
	for (int i = 0; i < num_thrds; i++)
	{
		delete thrd[i];
	}
	delete [] thrd;
}
