/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2003-2004 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "pdevs.h"
using namespace std;
USING_NAMESPACE_ADEVS

/**
For backwards compatibility with old config files that do not have the
INITIAL_MODEL_COUNT defined).
*/
#ifndef INITIAL_MODEL_COUNT
#define INITIAL_MODEL_COUNT 100
#endif

ADEVS_NAMESPACE::pdevs_coord::pdevs_coord(devsn* model, ADEVS_TIME_TYPE t_granule):
basic_sim(t_granule),
top_model(model),
imm(INITIAL_MODEL_COUNT),
sched(INITIAL_MODEL_COUNT),
io_pool(INITIAL_MODEL_COUNT,INITIAL_MODEL_COUNT)
{
	// Schedule all of the component models.  Computes global_tL during scheduling.
	global_tL = ADEVS_ZERO;
	// Models are scheduled recursively using the schedule() method
	schedule(top_model);
	// Set the time of next event
	global_tN = sched.minPriority();
}

void ADEVS_NAMESPACE::pdevs_coord::schedule(devsn* model)
{
	// Schedule the network executive
	sched.schedule(model->getNetExec(),model->getNetExec()->timeNext());
	// Update the global time of last event
	if (model->getNetExec()->timeLast() > global_tL) 
	{
		global_tL = model->getNetExec()->timeLast();
	}
	// Schedule each component model
	adevs_set<devs*> components;
	model->getNetExec()->getComponents(components);
	for (int i = 0; i < components.getSize(); i++)
	{
		// Atomic models are placed into the schedule
		atomic* amodel = components[i]->typeIsAtomic();
		if (amodel != NULL)
		{
			sched.schedule(amodel,amodel->timeNext());
			// Update the global time of last event
			if (amodel->timeLast() > global_tL)
			{
				global_tL = amodel->timeLast();
			}
		}
		// Network models are scheduled recursively
		else
		{
			schedule(components[i]->typeIsNetwork());
		}
	}
}

void ADEVS_NAMESPACE::pdevs_coord::computeInputOutput()
{
	// No imminent models, so no output
	if (sched.getSize() == 0) return;
	// Get all of the imminent models
	while (sched.minPriority() - global_tN <= t_granule)
	{
		atomic* imm_model = sched.getMinimum();
		sched.removeMinimum();
		if (imm_model->asim_support.active == false)
		{
			imm.add(imm_model);
			imm_model->asim_support.active = true;
		}
		if (sched.getSize() == 0)
		{
			break;
		}
	}
	// Remember the number of real imminents.  We use the fact that elements are
	// always added to the back of the 'bag' class to distinguish between imminent
	// models and models that are just undergoing an external transition.
	int num_imm = imm.getSize();
	// Compute I/O functions and route events for imminent models.  Save bags of output
	// values for garbage collection at the end of the cycle.
	for (int i = 0; i < num_imm; i++)
	{
		adevs_bag<PortValue>* output_values = io_pool.make_obj();
		// Make sure the output function sees the proper current time
		// (this is for backwards compatibility).
		ADEVS_TIME_TYPE t = imm[i]->timeCurrent();
		imm[i]->asim_support.t = imm[i]->timeNext();
		imm[i]->output_func(*output_values);
		// Restore the model's clock (no real advance in time takes place until
		// state transition function is computed).
		imm[i]->asim_support.t = t;
		// Send model outputs to their proper destination via the recursive routing
		// function.
		for (int j = 0; j < output_values->getSize(); j++)
		{
			route(imm[i]->bsim_support.parent,imm[i],output_values->get(j));
		}
		// Remember output values for later garbage collection.
		imm[i]->asim_support.y = output_values;
	}
}

void ADEVS_NAMESPACE::pdevs_coord::route(devsn* parent, devs* src, PortValue& pv)
{
	adevs_bag<EventReceiver>* recvs = recv_pool.make_obj();
	// Compute the set of receivers for this port value pair
	parent->getNetExec()->route(pv,src,*recvs);
	// Deliver the event to each of its targets
	PortValue pvx;
	pvx.value = pv.value;
	atomic* amodel = NULL;
	for (int i = 0; i < recvs->getSize(); i++)
	{
		// Check for self-influencing error condition
		if (src == recvs->get(i).model)
		{
			THROW_BASE_EXCEPTION("Model tried to influence self");
		}
		pvx.port = recvs->get(i).port;
		// if destination is an atomic model add the event to the IO bag for that model
		// and add model to end of list of imminents for this cycle
		amodel = recvs->get(i).model->typeIsAtomic();
		if (amodel != NULL)
		{
			if (amodel->asim_support.x == NULL)
			{
				amodel->asim_support.x = io_pool.make_obj();
				if (amodel->asim_support.active == false)
				{
					amodel->asim_support.active = true;
					imm.add(amodel);
				}
			}
			amodel->asim_support.x->add(pvx);
		}
		// otherwise if this is an external output of the top level model, add it
		// to the output table
		else if (recvs->get(i).model == top_model)
		{
			top_model_output.add(pvx);
		}
		// else if this is an external output from the parent model, route it
		// recursively
		else if (recvs->get(i).model == parent)
		{
			route(parent->bsim_support.parent,parent,pvx);
		}
		// otherwise, it is an input to a coupled model and should be routed recursively
		else
		{
			route(recvs->get(i).model->typeIsNetwork(),recvs->get(i).model,pvx);
		}
	}
	recvs->removeAll();
	recv_pool.destroy_obj(recvs);
}

void ADEVS_NAMESPACE::pdevs_coord::reschedule_atomic(atomic* model, bool force_passive)
{
	// Clean up the input bag for the model
	if (model->asim_support.x != NULL)
	{
		model->asim_support.x->removeAll();
		io_pool.destroy_obj(model->asim_support.x);
		model->asim_support.x = NULL;
	}
	// Clean up the output bag for the model
	if (model->asim_support.y != NULL)
	{
		model->gc_output(*(model->asim_support.y));
		model->asim_support.y->removeAll();
		io_pool.destroy_obj(model->asim_support.y);
		model->asim_support.y = NULL;
	}
	// Reschedule the model
	if (force_passive)
	{
		sched.schedule(model,ADEVS_INFINITY);
	}
	else
	{
		sched.schedule(model,model->timeNext());
	}
}

void ADEVS_NAMESPACE::pdevs_coord::compute_atomic_delta_func(atomic* model, 
ADEVS_TIME_TYPE t)
{
	adevs_bag<PortValue>* x = model->asim_support.x;
	model->asim_support.t = t;
	if (model->sigma() <= t_granule && x == NULL)
	{
		model->delta_int();
		model->asim_support.tL = t;
		model->asim_support.tN = t + model->ta();
	}
	else if (model->sigma() <= t_granule && x != NULL)
	{
		model->delta_conf(*x);
		model->asim_support.tL = t;
		model->asim_support.tN = t + model->ta();
	}
	else if (model->sigma() > t_granule && x != NULL)
	{
		model->delta_ext(model->elapsed(),*x);
		model->asim_support.tL = t;
		model->asim_support.tN = t + model->ta();
	}
	if (x != NULL)
	{
		x->removeAll();
		io_pool.destroy_obj(x);
		model->asim_support.x = NULL;
	}
}

void ADEVS_NAMESPACE::pdevs_coord::computeDeltaFunc(ADEVS_TIME_TYPE t)
{
	// index variables
	int i;
	// Check the requested time of next event for validity
	if (t < global_tL || t > global_tN + t_granule)
	{
		THROW_BASE_EXCEPTION("Invalid time of next event");
	}
	// If t is too early, we have no imminent models
	// (remember, computeInputOutput assumes t = tN)
	// so place all models imminent at tN back into the
	// schedule
	if (global_tN - t > t_granule)
	{
		for (i = 0; i < imm.getSize(); i++)
		{
			reschedule_atomic(imm[i]);
		}
		imm.removeAll();
	}
	// Route external input
	for (i = 0; i < top_model_input.getSize(); i++)
	{
		route(top_model,top_model,top_model_input[i]);
	}
	/**
	Compute next state for all imminent models.  If the model is a network executive,
	construct the new model and deleted model sets as well.  New models are added
	to the end of the immininent set and will be scheduled at the end of the simulation
	cycle.
	*/
	bool component_set_changed = false;
	for (i = 0; i < imm.getSize(); i++)
	{
		if (imm[i]->typeIsNetExec() != NULL)
		{
			nx_imm.add(imm[i]->typeIsNetExec());
			component_set_changed = 
				component_set_changed || imm[i]->typeIsNetExec()->componentSetChanged();
		}
		else
		{
			compute_atomic_delta_func(imm[i],t);
		}
	}
	/// Compute new states for network executive models and apply structure changes
	compute_structure_changes(component_set_changed,t);
	/// Set time of last event
	global_tL = t;
}

/**
Recursively removes a model from the simulator.
*/
void ADEVS_NAMESPACE::pdevs_coord::remove_model(devs* model)
{
	// If the model is atomic, remove it from the schedule, recover any
	// I/O bags being used by the model, and remove it from the imminent set.
	if (model->typeIsAtomic())
	{
		/// Remove it.  True flag forces removal rather than reschedule.
		reschedule_atomic(model->typeIsAtomic(),true);
		if (model->typeIsAtomic()->asim_support.active)
		{
			model->typeIsAtomic()->asim_support.active = false;
			imm.erase(model->typeIsAtomic());
		}
	}
	// Otherwise, get the components and remove them recursively
	else
	{
		adevs_set<devs*> components;
		model->typeIsNetwork()->getNetExec()->getComponents(components);
		for (int i = 0; i < components.getSize(); i++)
		{
			remove_model(components[i]);
		}
		remove_model(model->typeIsNetwork()->getNetExec());
	}
}

/**
Computes structure changes.  New models are initialized and scheduled.
Old models are removed from simulator and deleted.
*/
void ADEVS_NAMESPACE::pdevs_coord::compute_structure_changes(bool component_set_changed,
ADEVS_TIME_TYPE t) 
{
	// For loop index.  M$ VC++ 6 has strange scoping rules, so this has
	// to be done once, at the beginning, to avoid redeclaration errors.
	int i, j;
	// If no network executive changes its component set, then this is
	// very easy.  Just find the new states for the network executives
	if (!component_set_changed)
	{
		for (i = 0; i < nx_imm.getSize(); i++)
		{
			compute_atomic_delta_func(nx_imm[i],t);
		}
	}
	// Otherwise, we need to modify the simulator to accomodate model additions
	// and removals
	else
	{
		adevs_set<devs*> added, removed, prev, next;
		adevs_set<netExec*> safe_to_delete;
		// Compute state changes for all of the network executives.  The
		// sets of added and removed models are constructed here as well.
		for (i = 0; i < nx_imm.getSize(); i++)
		{
			// The safe set is used to avoid double deletes (e.g., as might
			// occur is two network executives change state, and one is the child
			// of another).
			nx_imm[i]->nsim_support.safe_set = &safe_to_delete;
			safe_to_delete.add(nx_imm[i]);
			// Get the structure change hint
			component_set_changed = nx_imm[i]->componentSetChanged();
			// Do structure change unless hint says not to do it
			if (component_set_changed)
			{
				// Get component set before the state transition
				nx_imm[i]->getComponents(prev);
			}
			// Compute the next state
			compute_atomic_delta_func(nx_imm[i],t);
			// Finish structure change according to the hint prior to the
			// state transition
			if (component_set_changed)
			{
				// Get the new component set
				nx_imm[i]->getComponents(next);
				// Set the parent of each of the new componet models
				for (j = 0;  j < next.getSize(); j++)
				{
					next[j]->bsim_support.parent = nx_imm[i]->getParent();
				}
				// Update the change sets
				added.append(next-prev);
				nx_imm[i]->nsim_support.garbage.append(prev-next);
				removed.append(nx_imm[i]->nsim_support.garbage);
				// Clean up the prev and next sets for the next round
				prev.removeAll();
				next.removeAll();
			}
		}
		// Any models that simply moved from one network or the other should be
		// left alone
		adevs_set<devs*> moved(removed.intersection(added));
		for (i = 0; i < moved.getSize(); i++)
		{
			added.erase(moved[i]);
			removed.erase(moved[i]);
			for (j = 0; j < nx_imm.getSize(); j++)
			{
				nx_imm[j]->nsim_support.garbage.erase(moved[i]);
			}
		}
		// Initialize and add all models in the added set (remember that the
		// parent field for these models has already been set correctly).
		for (i = 0; i < added.getSize(); i++)
		{
			// New atomic models need to be intialized and scheduled
			atomic* amodel = added[i]->typeIsAtomic();
			if (amodel != NULL)
			{
				amodel->asim_support.tL = t;
				amodel->asim_support.t = t;
				amodel->initialize(t);
				sched.schedule(amodel,amodel->timeNext());
			}
			// New network models are initialized and then scheduled recursively
			else
			{
				added[i]->initialize(t);
				schedule(added[i]->typeIsNetwork());
			}
		}
		// Remove all models in the removed set from the simulator.  This could
		// indirectly cause the removal of a model that was added in the previous step
		// and so it must be done second.
		for (i = 0; i < removed.getSize(); i++)
		{
			remove_model(removed[i]);
		}
		// Give network executives the chance to delete any removed components
		for (i = 0; i < nx_imm.getSize(); i++)
		{
			if (safe_to_delete.contains(nx_imm[i]))
			{
				nx_imm[i]->gc_models(nx_imm[i]->nsim_support.garbage);
				nx_imm[i]->nsim_support.garbage.removeAll();
				nx_imm[i]->nsim_support.safe_set = NULL;
			}
		}
	}
	// Done with the network execs, so clean up
	nx_imm.removeAll();
}

void ADEVS_NAMESPACE::pdevs_coord::endOfCycle()
{
	// Cleanup IO and reschedule all models that changed stated
	for (int i = 0; i < imm.getSize(); i++)
	{
		// Give models a chance to clean up their output values for this cycle
		if (imm[i]->asim_support.y != NULL)
		{
			imm[i]->gc_output(*(imm[i]->asim_support.y));
			imm[i]->asim_support.y->removeAll();
			io_pool.destroy_obj(imm[i]->asim_support.y);
			imm[i]->asim_support.y = NULL;
		}
		// Model is no longer active
		imm[i]->asim_support.active = false;
		// Schedule the models that changed state this cycle
		sched.schedule(imm[i],imm[i]->timeNext());
	}
	imm.removeAll();
	top_model_input.removeAll();
	top_model_output.removeAll();
	global_tN = sched.minPriority();
}

ADEVS_NAMESPACE::pdevs_coord::~pdevs_coord()
{
}
