/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2005 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "par_pdevs.h"
USING_NAMESPACE_ADEVS

#define IDLE 0
#define COMPUTE_IO 1
#define COMPUTE_STATE 2
#define COMPUTE_END 3
#define EXIT_THREAD 4

#ifdef USE_ADEVS_NAMESPACE
void* ADEVS_NAMESPACE::sim_thread_main(void* arg)
#else
void* sim_thread_main(void* arg)
#endif
{
	sim_thread* thr = static_cast<sim_thread*>(arg);
	// Enter the main simulation loop
	bool exit_loop = false;
	while (!exit_loop)
	{
		// Wait for a dispatch
		pthread_mutex_lock(&(thr->cond_mtx));
		while (thr->run_method == IDLE)
		{
			pthread_cond_wait(&(thr->run_cond),&(thr->cond_mtx));
		}
		// Execute the method
		if (thr->run_method == COMPUTE_IO)
		{
			thr->computeIO();
		}
		else if (thr->run_method == COMPUTE_STATE)
		{
			thr->computeNextState();
		}
		else if (thr->run_method == COMPUTE_END)
		{
			thr->computeEndOfCycle();
		}
		else
		{
			exit_loop = true;
		}
		// Signal that the method is done
		thr->run_method = IDLE;
		pthread_cond_broadcast(&(thr->run_cond));
		pthread_mutex_unlock(&(thr->cond_mtx));
	}
	return NULL;
}

ADEVS_NAMESPACE::sim_thread::sim_thread(int thrd_id, int num_thrds, ADEVS_TIME_TYPE t_granule):
t_granule(t_granule),
run_method(IDLE),
num_thrds(num_thrds),
thrd_id(thrd_id)
{
	x_thrd_io = new adevs_bag<model_io_t>[num_thrds];
	pthread_mutex_init(&cond_mtx,NULL);
	pthread_cond_init(&run_cond,NULL);
	pthread_create(&thr_hnd,NULL,sim_thread_main,(void*)this);
}

void ADEVS_NAMESPACE::sim_thread::initial_assignment(atomic* model)
{
	sched.schedule(model,model->timeNext());
	model->asim_support.owning_thr = (void*)this;
}

void ADEVS_NAMESPACE::sim_thread::activate_local(atomic* model, PortValue& pvx)
{
	if (model->asim_support.x == NULL)
	{
		model->asim_support.x = io_pool.make_obj();
		if (model->asim_support.active == false)
		{
			model->asim_support.active = true;
			imm.add(model);
		}
	}
	model->asim_support.x->add(pvx);
}

void ADEVS_NAMESPACE::sim_thread::activate(atomic* model, PortValue& pvx, int calling_thrd)
{
	model_io_t x;
	x.model = model;
	x.pv = pvx;
	x_thrd_io[calling_thrd].add(x);
}

void ADEVS_NAMESPACE::sim_thread::computeInputOutput(ADEVS_TIME_TYPE gvt)
{
	this->gvt = gvt;
	dispatch(COMPUTE_IO);
}

void ADEVS_NAMESPACE::sim_thread::computeDeltFunc()
{
	dispatch(COMPUTE_STATE);
}

void ADEVS_NAMESPACE::sim_thread::endOfCycle()
{
	dispatch(COMPUTE_END);
}

void ADEVS_NAMESPACE::sim_thread::wait()
{
	pthread_mutex_lock(&cond_mtx);
	while (run_method != IDLE)
	{
		pthread_cond_wait(&run_cond,&cond_mtx);
	}
	pthread_mutex_unlock(&cond_mtx);
}

ADEVS_NAMESPACE::sim_thread::~sim_thread()
{
	dispatch(EXIT_THREAD);
	pthread_join(thr_hnd,NULL);
	pthread_mutex_destroy(&cond_mtx);
	pthread_cond_destroy(&run_cond);
	delete [] x_thrd_io;
}

void ADEVS_NAMESPACE::sim_thread::dispatch(int method)
{
	pthread_mutex_lock(&cond_mtx);
	run_method = method;
	pthread_cond_broadcast(&run_cond);
	pthread_mutex_unlock(&cond_mtx);
}

void ADEVS_NAMESPACE::sim_thread::computeIO()
{
	if (sched.getSize() == 0) return;
	while (sched.minPriority() - gvt <= t_granule)
	{
		atomic* imm_model = sched.getMinimum();
		sched.removeMinimum();
		if (imm_model->asim_support.active == false)
		{
			imm_model->asim_support.active = true;
			imm.add(imm_model);
		}
		adevs_bag<PortValue>* output_values = io_pool.make_obj();
		ADEVS_TIME_TYPE t = imm_model->timeCurrent();
		imm_model->asim_support.t = imm_model->timeNext();
		imm_model->output_func(*output_values);
		imm_model->asim_support.t = t;
		for (int i = 0; i < output_values->getSize(); i++)
		{
			route(imm_model->bsim_support.parent,imm_model,output_values->get(i));
		}
		imm_model->asim_support.y = output_values;
	}
}

void ADEVS_NAMESPACE::sim_thread::route(devsn* parent, devs* src, PortValue& pv)
{
	/**
	Because injection and output is not supported, any output directed at the "top model" is
	simply discarded.
	*/
	if (parent == NULL)
	{
		return;
	}
	/// Otherwise, route the event
	adevs_bag<EventReceiver>* recvs = recv_pool.make_obj();
	// Compute the set of receivers for this port value pair.  This requires that
	// the network executives' route() method be reentrant, otherwise this will fail.
	parent->getNetExec()->route(pv,src,*recvs);
   // Deliver the event to each of its targets
   for (int i = 0; i < recvs->getSize(); i++)
   {
      // Check for self-influencing error condition
      if (src == recvs->get(i).model)
      {
         THROW_BASE_EXCEPTION("Model tried to influence self");
      }
      PortValue pvx(recvs->get(i).port,pv.value);
      // if destination is an atomic model, then add the event to the IO bag for that model
      // and add model to end of list of imminents for this cycle
		atomic* amodel = recvs->get(i).model->typeIsAtomic(); 
      if (amodel != NULL)
      {
			// If the model is local to the thread.
			if (amodel->asim_support.owning_thr == this)
			{
				activate_local(amodel,pvx);
			}
			// Otherwise, this will require cross thread communication
			else
			{
				sim_thread* owner = static_cast<sim_thread*>(amodel->asim_support.owning_thr);
				owner->activate(amodel,pvx,thrd_id);
			}
		}
		// else if this is an external output from the parent model, route it
		// recursively
		else if (recvs->get(i).model == parent)
		{
			route(parent->bsim_support.parent,parent,pvx);
		}
		// otherwise, it is an input to a coupled model and should be routed recursively
		else
		{
			route(recvs->get(i).model->typeIsNetwork(),recvs->get(i).model,pvx);
		}
	}
	recvs->removeAll();
	recv_pool.destroy_obj(recvs);
}
		
void ADEVS_NAMESPACE::sim_thread::computeNextState()
{
	// Apply all of the input received from outside of this thread
	for (int j = 0; j < num_thrds; j++)
	{
		for (int i = 0; i < x_thrd_io[j].getSize(); i++)
		{
			activate_local(x_thrd_io[j][i].model,x_thrd_io[j][i].pv);
		}
		x_thrd_io[j].removeAll();
	}
	// Compute state changes for all local models
	for (int i = 0; i < imm.getSize(); i++)
	{
		atomic* model = imm[i];
		if (model->typeIsNetExec() != NULL)
		{
			THROW_BASE_EXCEPTION("parallel simulator does not support dynamic structure models");
		}
		adevs_bag<PortValue>* x = model->asim_support.x;
		model->asim_support.t = gvt;
		if (model->sigma() <= t_granule && x == NULL)
		{
			model->delta_int();
			model->asim_support.tL = gvt;
			model->asim_support.tN = gvt + model->ta();
		}
		else if (model->sigma() <= t_granule && x != NULL)
		{
			model->delta_conf(*x);
			model->asim_support.tL = gvt;
			model->asim_support.tN = gvt + model->ta();
		}
		else if (model->sigma() > t_granule && x != NULL)
		{
			model->delta_ext(model->elapsed(),*x);
			model->asim_support.tL = gvt;
			model->asim_support.tN = gvt + model->ta();
		}
		if (x != NULL)
		{
			x->removeAll();
			io_pool.destroy_obj(x);
			model->asim_support.x = NULL;
		}
	}
}

void ADEVS_NAMESPACE::sim_thread::computeEndOfCycle()
{
	// Cleanup IO and reschedule all models that changed stated
	for (int i = 0; i < imm.getSize(); i++)
	{
		// Give models a chance to clean up their output values for this cycle
		if (imm[i]->asim_support.y != NULL)
		{
			imm[i]->gc_output(*(imm[i]->asim_support.y));
			imm[i]->asim_support.y->removeAll();
			io_pool.destroy_obj(imm[i]->asim_support.y);
			imm[i]->asim_support.y = NULL;
		}
		// Model is no longer active
		imm[i]->asim_support.active = false;
		// Schedule the models that changed state this cycle and that we own
		sched.schedule(imm[i],imm[i]->timeNext());
	}
	imm.removeAll();
}

